# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import xml.parsers.expat

import xmltodict

from swh.indexer.codemeta import CROSSWALK_TABLE, SCHEMA_URI
from .base import DictMapping, SingleFileMapping


class MavenMapping(DictMapping, SingleFileMapping):
    """
    dedicated class for Maven (pom.xml) mapping and translation
    """
    name = 'maven'
    filename = b'pom.xml'
    mapping = CROSSWALK_TABLE['Java (Maven)']
    string_fields = ['name', 'version', 'description', 'email']

    def translate(self, content):
        try:
            d = xmltodict.parse(content).get('project') or {}
        except xml.parsers.expat.ExpatError:
            self.log.warning('Error parsing XML from %s', self.log_suffix)
            return None
        except UnicodeDecodeError:
            self.log.warning('Error unidecoding XML from %s', self.log_suffix)
            return None
        except (LookupError, ValueError):
            # unknown encoding or multi-byte encoding
            self.log.warning('Error detecting XML encoding from %s',
                             self.log_suffix)
            return None
        metadata = self._translate_dict(d, normalize=False)
        metadata[SCHEMA_URI+'codeRepository'] = self.parse_repositories(d)
        metadata[SCHEMA_URI+'license'] = self.parse_licenses(d)
        return self.normalize_translation(metadata)

    _default_repository = {'url': 'https://repo.maven.apache.org/maven2/'}

    def parse_repositories(self, d):
        """https://maven.apache.org/pom.html#Repositories

        >>> import xmltodict
        >>> from pprint import pprint
        >>> d = xmltodict.parse('''
        ... <repositories>
        ...   <repository>
        ...     <id>codehausSnapshots</id>
        ...     <name>Codehaus Snapshots</name>
        ...     <url>http://snapshots.maven.codehaus.org/maven2</url>
        ...     <layout>default</layout>
        ...   </repository>
        ... </repositories>
        ... ''')
        >>> MavenMapping().parse_repositories(d)
        """
        repositories = d.get('repositories')
        if not repositories:
            results = [self.parse_repository(d, self._default_repository)]
        elif isinstance(repositories, dict):
            repositories = repositories.get('repository') or []
            if not isinstance(repositories, list):
                repositories = [repositories]
            results = [self.parse_repository(d, repo)
                       for repo in repositories]
        else:
            results = []
        return [res for res in results if res] or None

    def parse_repository(self, d, repo):
        if not isinstance(repo, dict):
            return
        if repo.get('layout', 'default') != 'default':
            return  # TODO ?
        url = repo.get('url')
        group_id = d.get('groupId')
        artifact_id = d.get('artifactId')
        if (isinstance(url, str) and isinstance(group_id, str)
                and isinstance(artifact_id, str)):
            repo = os.path.join(url, *group_id.split('.'), artifact_id)
            return {"@id": repo}

    def normalize_groupId(self, id_):
        """https://maven.apache.org/pom.html#Maven_Coordinates

        >>> MavenMapping().normalize_groupId('org.example')
        {'@id': 'org.example'}
        """
        if isinstance(id_, str):
            return {"@id": id_}

    def parse_licenses(self, d):
        """https://maven.apache.org/pom.html#Licenses

        >>> import xmltodict
        >>> import json
        >>> d = xmltodict.parse('''
        ... <licenses>
        ...   <license>
        ...     <name>Apache License, Version 2.0</name>
        ...     <url>https://www.apache.org/licenses/LICENSE-2.0.txt</url>
        ...   </license>
        ... </licenses>
        ... ''')
        >>> print(json.dumps(d, indent=4))
        {
            "licenses": {
                "license": {
                    "name": "Apache License, Version 2.0",
                    "url": "https://www.apache.org/licenses/LICENSE-2.0.txt"
                }
            }
        }
        >>> MavenMapping().parse_licenses(d)
        [{'@id': 'https://www.apache.org/licenses/LICENSE-2.0.txt'}]

        or, if there are more than one license:

        >>> import xmltodict
        >>> from pprint import pprint
        >>> d = xmltodict.parse('''
        ... <licenses>
        ...   <license>
        ...     <name>Apache License, Version 2.0</name>
        ...     <url>https://www.apache.org/licenses/LICENSE-2.0.txt</url>
        ...   </license>
        ...   <license>
        ...     <name>MIT License</name>
        ...     <url>https://opensource.org/licenses/MIT</url>
        ...   </license>
        ... </licenses>
        ... ''')
        >>> pprint(MavenMapping().parse_licenses(d))
        [{'@id': 'https://www.apache.org/licenses/LICENSE-2.0.txt'},
         {'@id': 'https://opensource.org/licenses/MIT'}]
        """

        licenses = d.get('licenses')
        if not isinstance(licenses, dict):
            return
        licenses = licenses.get('license')
        if isinstance(licenses, dict):
            licenses = [licenses]
        elif not isinstance(licenses, list):
            return
        return [{"@id": license['url']}
                for license in licenses
                if isinstance(license, dict)
                and isinstance(license.get('url'), str)] or None



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import email.parser
import email.policy
import itertools

from swh.indexer.codemeta import CROSSWALK_TABLE, SCHEMA_URI
from .base import DictMapping, SingleFileMapping


_normalize_pkginfo_key = str.lower


class LinebreakPreservingEmailPolicy(email.policy.EmailPolicy):
    def header_fetch_parse(self, name, value):
        if hasattr(value, 'name'):
            return value
        value = value.replace('\n        ', '\n')
        return self.header_factory(name, value)


class PythonPkginfoMapping(DictMapping, SingleFileMapping):
    """Dedicated class for Python's PKG-INFO mapping and translation.

    https://www.python.org/dev/peps/pep-0314/"""
    name = 'pkg-info'
    filename = b'PKG-INFO'
    mapping = {_normalize_pkginfo_key(k): v
               for (k, v) in CROSSWALK_TABLE['Python PKG-INFO'].items()}
    string_fields = ['name', 'version', 'description', 'summary',
                     'author', 'author-email']

    _parser = email.parser.BytesHeaderParser(
        policy=LinebreakPreservingEmailPolicy())

    def translate(self, content):
        msg = self._parser.parsebytes(content)
        d = {}
        for (key, value) in msg.items():
            key = _normalize_pkginfo_key(key)
            if value != 'UNKNOWN':
                d.setdefault(key, []).append(value)
        metadata = self._translate_dict(d, normalize=False)
        if SCHEMA_URI+'author' in metadata or SCHEMA_URI+'email' in metadata:
            metadata[SCHEMA_URI+'author'] = {
                '@list': [{
                    '@type': SCHEMA_URI+'Person',
                    SCHEMA_URI+'name':
                        metadata.pop(SCHEMA_URI+'author', [None])[0],
                    SCHEMA_URI+'email':
                        metadata.pop(SCHEMA_URI+'email', [None])[0],
                }]
            }
        return self.normalize_translation(metadata)

    def normalize_home_page(self, urls):
        return [{'@id': url} for url in urls]

    def normalize_keywords(self, keywords):
        return list(itertools.chain.from_iterable(
            s.split(' ') for s in keywords))

    def normalize_license(self, licenses):
        return [{'@id': license} for license in licenses]



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
import unittest
from unittest.mock import patch

import pytest

import swh.indexer.ctags
from swh.indexer.ctags import (
    CtagsIndexer, run_ctags
)

from swh.indexer.tests.utils import (
    CommonContentIndexerTest,
    SHA1_TO_CTAGS, BASE_TEST_CONFIG,
    OBJ_STORAGE_DATA, fill_storage, fill_obj_storage,
    filter_dict,
)


class BasicTest(unittest.TestCase):
    @patch("swh.indexer.ctags.subprocess")
    def test_run_ctags(self, mock_subprocess):
        """Computing licenses from a raw content should return results

        """
        output0 = """
{"name":"defun","kind":"function","line":1,"language":"scheme"}
{"name":"name","kind":"symbol","line":5,"language":"else"}"""
        output1 = """
{"name":"let","kind":"var","line":10,"language":"something"}"""

        expected_result0 = [
            {
                'name': 'defun',
                'kind': 'function',
                'line': 1,
                'lang': 'scheme'
            },
            {
                'name': 'name',
                'kind': 'symbol',
                'line': 5,
                'lang': 'else'
            }
        ]

        expected_result1 = [
            {
                'name': 'let',
                'kind': 'var',
                'line': 10,
                'lang': 'something'
            }
        ]
        for path, lang, intermediary_result, expected_result in [
                (b'some/path', 'lisp', output0, expected_result0),
                (b'some/path/2', 'markdown', output1, expected_result1)
        ]:
            mock_subprocess.check_output.return_value = intermediary_result
            actual_result = list(run_ctags(path, lang=lang))
            self.assertEqual(actual_result, expected_result)


class InjectCtagsIndexer:
    """Override ctags computations.

    """
    def compute_ctags(self, path, lang):
        """Inject fake ctags given path (sha1 identifier).

        """
        return {
            'lang': lang,
            **SHA1_TO_CTAGS.get(path)
        }


CONFIG = {
    **BASE_TEST_CONFIG,
    'tools': {
        'name': 'universal-ctags',
        'version': '~git7859817b',
        'configuration': {
            'command_line': '''ctags --fields=+lnz --sort=no '''
            ''' --links=no <filepath>''',
            'max_content_size': 1000,
        },
    },
    'languages': {
        'python': 'python',
        'haskell': 'haskell',
        'bar': 'bar',
    },
    'workdir': '/tmp',
}


class TestCtagsIndexer(CommonContentIndexerTest, unittest.TestCase):
    """Ctags indexer test scenarios:

    - Known sha1s in the input list have their data indexed
    - Unknown sha1 in the input list are not indexed

    """

    legacy_get_format = True

    def get_indexer_results(self, ids):
        yield from self.idx_storage.content_ctags_get(ids)

    def setUp(self):
        super().setUp()
        self.indexer = CtagsIndexer(config=CONFIG)
        self.indexer.catch_exceptions = False
        self.idx_storage = self.indexer.idx_storage
        fill_storage(self.indexer.storage)
        fill_obj_storage(self.indexer.objstorage)

        # Prepare test input
        self.id0 = '01c9379dfc33803963d07c1ccc748d3fe4c96bb5'
        self.id1 = 'd4c647f0fc257591cc9ba1722484229780d1c607'
        self.id2 = '688a5ef812c53907562fe379d4b3851e69c7cb15'

        tool = {k.replace('tool_', ''): v
                for (k, v) in self.indexer.tool.items()}

        self.expected_results = {
            self.id0: {
                'id': self.id0,
                'tool': tool,
                **SHA1_TO_CTAGS[self.id0][0],
            },
            self.id1: {
                'id': self.id1,
                'tool': tool,
                **SHA1_TO_CTAGS[self.id1][0],
            },
            self.id2: {
                'id': self.id2,
                'tool': tool,
                **SHA1_TO_CTAGS[self.id2][0],
            }
        }

        self._set_mocks()

    def _set_mocks(self):
        def find_ctags_for_content(raw_content):
            for (sha1, ctags) in SHA1_TO_CTAGS.items():
                if OBJ_STORAGE_DATA[sha1] == raw_content:
                    return ctags
            else:
                raise ValueError(('%r not found in objstorage, can\'t mock '
                                  'its ctags.') % raw_content)

        def fake_language(raw_content, *args, **kwargs):
            ctags = find_ctags_for_content(raw_content)
            return {'lang': ctags[0]['lang']}
        self._real_compute_language = swh.indexer.ctags.compute_language
        swh.indexer.ctags.compute_language = fake_language

        def fake_check_output(cmd, *args, **kwargs):
            print(cmd)
            id_ = cmd[-1].split('/')[-1]
            return '\n'.join(
                json.dumps({'language': ctag['lang'], **ctag})
                for ctag in SHA1_TO_CTAGS[id_])
        self._real_check_output = swh.indexer.ctags.subprocess.check_output
        swh.indexer.ctags.subprocess.check_output = fake_check_output

    def tearDown(self):
        swh.indexer.ctags.compute_language = self._real_compute_language
        swh.indexer.ctags.subprocess.check_output = self._real_check_output
        super().tearDown()


def test_ctags_w_no_tool():
    with pytest.raises(ValueError):
        CtagsIndexer(config=filter_dict(CONFIG, 'tools'))



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging

from swh.core.utils import grouper
from swh.scheduler.utils import create_task_dict


MAX_ORIGINS_PER_TASK = 100


def process_journal_objects(messages, *, scheduler, task_names):
    """Worker function for `JournalClient.process(worker_fn)`, after
    currification of `scheduler` and `task_names`."""
    assert set(messages) == {'origin_visit'}, set(messages)
    process_origin_visits(messages['origin_visit'], scheduler, task_names)


def process_origin_visits(visits, scheduler, task_names):
    task_dicts = []
    logging.debug('processing origin visits %r', visits)
    if task_names.get('origin_metadata'):
        visits = [visit for visit in visits if visit['status'] == 'full']
        visit_batches = grouper(visits, MAX_ORIGINS_PER_TASK)
        for visit_batch in visit_batches:
            visit_urls = []
            for visit in visit_batch:
                if isinstance(visit['origin'], str):
                    visit_urls.append(visit['origin'])
                else:
                    visit_urls.append(visit['origin']['url'])
            task_dicts.append(create_task_dict(
                task_names['origin_metadata'],
                'oneshot',
                visit_urls,
                policy_update='update-dups',
            ))

    if task_dicts:
        scheduler.create_tasks(task_dicts)



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import itertools

from collections import defaultdict

from swh.core import utils
from swh.core.config import SWHConfig
from swh.model import hashutil
from swh.objstorage import get_objstorage
from swh.objstorage.exc import ObjNotFoundError
from swh.storage import get_storage


class RecomputeChecksums(SWHConfig):
    """Class in charge of (re)computing content's hashes.

    Hashes to compute are defined across 2 configuration options:

    compute_checksums ([str])
      list of hash algorithms that
      py:func:`swh.model.hashutil.MultiHash.from_data` function should
      be able to deal with. For variable-length checksums, a desired
      checksum length should also be provided. Their format is
      <algorithm's name>:<variable-length> e.g: blake2:512

    recompute_checksums (bool)
      a boolean to notify that we also want to recompute potential existing
      hashes specified in compute_checksums.  Default to False.

    """
    DEFAULT_CONFIG = {
        # The storage to read from or update metadata to
        'storage': ('dict', {
            'cls': 'remote',
            'args': {
              'url': 'http://localhost:5002/'
            },
        }),
        # The objstorage to read contents' data from
        'objstorage': ('dict', {
            'cls': 'pathslicing',
            'args': {
                'root': '/srv/softwareheritage/objects',
                'slicing': '0:2/2:4/4:6',
            },
        }),
        # the set of checksums that should be computed.
        # Examples: 'sha1_git', 'blake2b512', 'blake2s256'
        'compute_checksums': (
            'list[str]', []),
        # whether checksums that already exist in the DB should be
        # recomputed/updated or left untouched
        'recompute_checksums': ('bool', False),
        # Number of contents to retrieve blobs at the same time
        'batch_size_retrieve_content': ('int', 10),
        # Number of contents to update at the same time
        'batch_size_update': ('int', 100),
    }

    CONFIG_BASE_FILENAME = 'indexer/rehash'

    def __init__(self):
        self.config = self.parse_config_file()
        self.storage = get_storage(**self.config['storage'])
        self.objstorage = get_objstorage(**self.config['objstorage'])
        self.compute_checksums = self.config['compute_checksums']
        self.recompute_checksums = self.config[
            'recompute_checksums']
        self.batch_size_retrieve_content = self.config[
            'batch_size_retrieve_content']
        self.batch_size_update = self.config[
            'batch_size_update']
        self.log = logging.getLogger('swh.indexer.rehash')

        if not self.compute_checksums:
            raise ValueError('Checksums list should not be empty.')

    def _read_content_ids(self, contents):
        """Read the content identifiers from the contents.

        """
        for c in contents:
            h = c['sha1']
            if isinstance(h, str):
                h = hashutil.hash_to_bytes(h)

            yield h

    def get_new_contents_metadata(self, all_contents):
        """Retrieve raw contents and compute new checksums on the
           contents. Unknown or corrupted contents are skipped.

        Args:
            all_contents ([dict]): List of contents as dictionary with
              the necessary primary keys
            checksum_algorithms ([str]): List of checksums to compute

        Yields:
            tuple: tuple of (content to update, list of checksums computed)

        """
        content_ids = self._read_content_ids(all_contents)
        for contents in utils.grouper(content_ids,
                                      self.batch_size_retrieve_content):
            contents_iter = itertools.tee(contents, 2)
            try:
                content_metadata = self.storage.content_get_metadata(
                    [s for s in contents_iter[0]])
            except Exception:
                self.log.exception(
                    'Problem when reading contents metadata.')
                continue

            for content in content_metadata:
                # Recompute checksums provided in compute_checksums options
                if self.recompute_checksums:
                    checksums_to_compute = list(self.compute_checksums)
                else:
                    # Compute checksums provided in compute_checksums
                    # options not already defined for that content
                    checksums_to_compute = [h for h in self.compute_checksums
                                            if not content.get(h)]

                if not checksums_to_compute:  # Nothing to recompute
                    continue

                try:
                    raw_content = self.objstorage.get(content['sha1'])
                except ObjNotFoundError:
                    self.log.warning('Content %s not found in objstorage!' %
                                     content['sha1'])
                    continue

                content_hashes = hashutil.MultiHash.from_data(
                    raw_content, hash_names=checksums_to_compute).digest()
                content.update(content_hashes)
                yield content, checksums_to_compute

    def run(self, contents):
        """Given a list of content:

          - (re)compute a given set of checksums on contents available in our
            object storage
          - update those contents with the new metadata

          Args:
              contents (dict): contents as dictionary with necessary keys.
                  key present in such dictionary should be the ones defined in
                  the 'primary_key' option.

        """
        for data in utils.grouper(
                self.get_new_contents_metadata(contents),
                self.batch_size_update):

            groups = defaultdict(list)
            for content, keys_to_update in data:
                keys = ','.join(keys_to_update)
                groups[keys].append(content)

            for keys_to_update, contents in groups.items():
                keys = keys_to_update.split(',')
                try:
                    self.storage.content_update(contents,
                                                keys=keys)
                except Exception:
                    self.log.exception('Problem during update.')
                    continue



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import timedelta
from unittest.mock import patch

import pytest

from swh.objstorage import get_objstorage
from swh.scheduler.tests.conftest import *  # noqa
from swh.storage import get_storage
from swh.indexer.storage import get_indexer_storage

from .utils import fill_storage, fill_obj_storage


TASK_NAMES = ['revision_intrinsic_metadata', 'origin_intrinsic_metadata']


@pytest.fixture
def indexer_scheduler(swh_scheduler):
    for taskname in TASK_NAMES:
        swh_scheduler.create_task_type({
            'type': taskname,
            'description': 'The {} indexer testing task'.format(taskname),
            'backend_name': 'swh.indexer.tests.tasks.{}'.format(taskname),
            'default_interval': timedelta(days=1),
            'min_interval': timedelta(hours=6),
            'max_interval': timedelta(days=12),
            'num_retries': 3,
        })
    return swh_scheduler


@pytest.fixture
def idx_storage():
    """An instance of in-memory indexer storage that gets injected into all
    indexers classes.

    """
    idx_storage = get_indexer_storage('memory', {})
    with patch('swh.indexer.storage.in_memory.IndexerStorage') \
            as idx_storage_mock:
        idx_storage_mock.return_value = idx_storage
        yield idx_storage


@pytest.fixture
def storage():
    """An instance of in-memory storage that gets injected into all indexers
       classes.

    """
    storage = get_storage('memory')
    fill_storage(storage)
    with patch('swh.storage.in_memory.InMemoryStorage') as storage_mock:
        storage_mock.return_value = storage
        yield storage


@pytest.fixture
def obj_storage():
    """An instance of in-memory objstorage that gets injected into all indexers
    classes.

    """
    objstorage = get_objstorage('memory', {})
    fill_obj_storage(objstorage)
    with patch.dict('swh.objstorage._STORAGE_CLASSES',
                    {'memory': lambda: objstorage}):
        yield objstorage


@pytest.fixture(scope='session')  # type: ignore  # expected redefinition
def celery_includes():
    return [
        'swh.indexer.tests.tasks',
        'swh.indexer.tasks',
    ]



import swh.indexer
from os import path

from celery.contrib.testing.worker import start_worker
import celery.contrib.testing.tasks  # noqa

from swh.scheduler.celery_backend.config import app

__all__ = ['start_worker_thread']

SQL_DIR = path.join(path.dirname(swh.indexer.__file__), 'sql')


def start_worker_thread():
    return start_worker(app)



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.core.api import RPCClient

from swh.storage.exc import StorageAPIError

from ..interface import IndexerStorageInterface


class RemoteStorage(RPCClient):
    """Proxy to a remote storage API"""

    backend_class = IndexerStorageInterface
    api_exception = StorageAPIError



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json

from swh.indexer.codemeta import CODEMETA_TERMS
from swh.indexer.codemeta import expand
from .base import SingleFileMapping


class CodemetaMapping(SingleFileMapping):
    """
    dedicated class for CodeMeta (codemeta.json) mapping and translation
    """
    name = 'codemeta'
    filename = b'codemeta.json'
    string_fields = None

    @classmethod
    def supported_terms(cls):
        return [term for term in CODEMETA_TERMS if not term.startswith('@')]

    def translate(self, content):
        try:
            return self.normalize_translation(expand(
                json.loads(content.decode())))
        except Exception:
            return None



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import re
import click
import logging

from swh.indexer.indexer import OriginIndexer


class OriginHeadIndexer(OriginIndexer):
    """Origin-level indexer.

    This indexer is in charge of looking up the revision that acts as the
    "head" of an origin.

    In git, this is usually the commit pointed to by the 'master' branch."""

    USE_TOOLS = False

    def persist_index_computations(self, results, policy_update):
        """Do nothing. The indexer's results are not persistent, they
        should only be piped to another indexer."""
        pass

    # Dispatch

    def index(self, origin_url):
        latest_visit = self.storage.origin_visit_get_latest(
            origin_url, allowed_statuses=['full'], require_snapshot=True)
        if latest_visit is None:
            return None
        latest_snapshot = self.storage.snapshot_get(latest_visit['snapshot'])
        method = getattr(
            self, '_try_get_%s_head' % latest_visit['type'],
            self._try_get_head_generic)

        rev_id = method(latest_snapshot)
        if rev_id is not None:
            return {
                'origin_url': origin_url,
                'revision_id': rev_id,
                }

        # could not find a head revision
        return None

    # Tarballs

    _archive_filename_re = re.compile(
            rb'^'
            rb'(?P<pkgname>.*)[-_]'
            rb'(?P<version>[0-9]+(\.[0-9])*)'
            rb'(?P<preversion>[-+][a-zA-Z0-9.~]+?)?'
            rb'(?P<extension>(\.[a-zA-Z0-9]+)+)'
            rb'$')

    @classmethod
    def _parse_version(cls, filename):
        """Extracts the release version from an archive filename,
        to get an ordering whose maximum is likely to be the last
        version of the software

        >>> OriginHeadIndexer._parse_version(b'foo')
        (-inf,)
        >>> OriginHeadIndexer._parse_version(b'foo.tar.gz')
        (-inf,)
        >>> OriginHeadIndexer._parse_version(b'gnu-hello-0.0.1.tar.gz')
        (0, 0, 1, 0)
        >>> OriginHeadIndexer._parse_version(b'gnu-hello-0.0.1-beta2.tar.gz')
        (0, 0, 1, -1, 'beta2')
        >>> OriginHeadIndexer._parse_version(b'gnu-hello-0.0.1+foobar.tar.gz')
        (0, 0, 1, 1, 'foobar')
        """
        res = cls._archive_filename_re.match(filename)
        if res is None:
            return (float('-infinity'),)
        version = [int(n) for n in res.group('version').decode().split('.')]
        if res.group('preversion') is None:
            version.append(0)
        else:
            preversion = res.group('preversion').decode()
            if preversion.startswith('-'):
                version.append(-1)
                version.append(preversion[1:])
            elif preversion.startswith('+'):
                version.append(1)
                version.append(preversion[1:])
            else:
                assert False, res.group('preversion')
        return tuple(version)

    def _try_get_ftp_head(self, snapshot):
        archive_names = list(snapshot['branches'])
        max_archive_name = max(archive_names, key=self._parse_version)
        r = self._try_resolve_target(snapshot['branches'], max_archive_name)
        return r

    # Generic

    def _try_get_head_generic(self, snapshot):
        # Works on 'deposit', 'pypi', and VCSs.
        try:
            branches = snapshot['branches']
        except KeyError:
            return None
        else:
            return (
                    self._try_resolve_target(branches, b'HEAD') or
                    self._try_resolve_target(branches, b'master')
                    )

    def _try_resolve_target(self, branches, target_name):
        try:
            target = branches[target_name]
            if target is None:
                return None
            while target['target_type'] == 'alias':
                target = branches[target['target']]
                if target is None:
                    return None

            if target['target_type'] == 'revision':
                return target['target']
            elif target['target_type'] == 'content':
                return None  # TODO
            elif target['target_type'] == 'directory':
                return None  # TODO
            elif target['target_type'] == 'release':
                return None  # TODO
            else:
                assert False
        except KeyError:
            return None


@click.command()
@click.option('--origins', '-i',
              help='Origins to lookup, in the "type+url" format',
              multiple=True)
def main(origins):
    rev_metadata_indexer = OriginHeadIndexer()
    rev_metadata_indexer.run(origins)


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    main()



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import bisect
from collections import defaultdict, Counter
import itertools
import json
import operator
import math
import re

from . import MAPPING_NAMES

SHA1_DIGEST_SIZE = 160


def _transform_tool(tool):
    return {
        'id': tool['id'],
        'name': tool['tool_name'],
        'version': tool['tool_version'],
        'configuration': tool['tool_configuration'],
    }


class SubStorage:
    """Implements common missing/get/add logic for each indexer type."""
    def __init__(self, tools):
        self._tools = tools
        self._sorted_ids = []
        self._data = {}  # map (id_, tool_id) -> metadata_dict
        self._tools_per_id = defaultdict(set)  # map id_ -> Set[tool_id]

    def missing(self, ids):
        """List data missing from storage.

        Args:
            data (iterable): dictionaries with keys:

                - **id** (bytes): sha1 identifier
                - **indexer_configuration_id** (int): tool used to compute
                  the results

        Yields:
            missing sha1s

        """
        for id_ in ids:
            tool_id = id_['indexer_configuration_id']
            id_ = id_['id']
            if tool_id not in self._tools_per_id.get(id_, set()):
                yield id_

    def get(self, ids):
        """Retrieve data per id.

        Args:
            ids (iterable): sha1 checksums

        Yields:
            dict: dictionaries with the following keys:

              - **id** (bytes)
              - **tool** (dict): tool used to compute metadata
              - arbitrary data (as provided to `add`)

        """
        for id_ in ids:
            for tool_id in self._tools_per_id.get(id_, set()):
                key = (id_, tool_id)
                yield {
                    'id': id_,
                    'tool': _transform_tool(self._tools[tool_id]),
                    **self._data[key],
                }

    def get_all(self):
        yield from self.get(self._sorted_ids)

    def get_range(self, start, end, indexer_configuration_id, limit):
        """Retrieve data within range [start, end] bound by limit.

        Args:
            **start** (bytes): Starting identifier range (expected smaller
                           than end)
            **end** (bytes): Ending identifier range (expected larger
                             than start)
            **indexer_configuration_id** (int): The tool used to index data
            **limit** (int): Limit result

        Raises:
            ValueError for limit to None

        Returns:
            a dict with keys:
            - **ids** [bytes]: iterable of content ids within the range.
            - **next** (Optional[bytes]): The next range of sha1 starts at
                                          this sha1 if any

        """
        if limit is None:
            raise ValueError('Development error: limit should not be None')
        from_index = bisect.bisect_left(self._sorted_ids, start)
        to_index = bisect.bisect_right(self._sorted_ids, end, lo=from_index)
        if to_index - from_index >= limit:
            return {
                'ids': self._sorted_ids[from_index:from_index+limit],
                'next': self._sorted_ids[from_index+limit],
            }
        else:
            return {
                'ids': self._sorted_ids[from_index:to_index],
                'next': None,
                }

    def add(self, data, conflict_update):
        """Add data not present in storage.

        Args:
            data (iterable): dictionaries with keys:

              - **id**: sha1
              - **indexer_configuration_id**: tool used to compute the
                results
              - arbitrary data

            conflict_update (bool): Flag to determine if we want to overwrite
              (true) or skip duplicates (false)

        """
        data = list(data)
        if len({x['id'] for x in data}) < len(data):
            # For "exception-compatibility" with the pgsql backend
            raise ValueError('The same id is present more than once.')
        for item in data:
            item = item.copy()
            tool_id = item.pop('indexer_configuration_id')
            id_ = item.pop('id')
            data = item
            if not conflict_update and \
                    tool_id in self._tools_per_id.get(id_, set()):
                # Duplicate, should not be updated
                continue
            key = (id_, tool_id)
            self._data[key] = data
            self._tools_per_id[id_].add(tool_id)
            if id_ not in self._sorted_ids:
                bisect.insort(self._sorted_ids, id_)

    def add_merge(self, new_data, conflict_update, merged_key):
        for new_item in new_data:
            id_ = new_item['id']
            tool_id = new_item['indexer_configuration_id']
            if conflict_update:
                all_subitems = []
            else:
                existing = list(self.get([id_]))
                all_subitems = [
                    old_subitem
                    for existing_item in existing
                    if existing_item['tool']['id'] == tool_id
                    for old_subitem in existing_item[merged_key]
                ]
            for new_subitem in new_item[merged_key]:
                if new_subitem not in all_subitems:
                    all_subitems.append(new_subitem)
            self.add([
                {
                    'id': id_,
                    'indexer_configuration_id': tool_id,
                    merged_key: all_subitems,
                }
            ], conflict_update=True)
            if id_ not in self._sorted_ids:
                bisect.insort(self._sorted_ids, id_)

    def delete(self, entries):
        for entry in entries:
            (id_, tool_id) = (entry['id'], entry['indexer_configuration_id'])
            key = (id_, tool_id)
            if tool_id in self._tools_per_id[id_]:
                self._tools_per_id[id_].remove(tool_id)
            if key in self._data:
                del self._data[key]


class IndexerStorage:
    """In-memory SWH indexer storage."""

    def __init__(self):
        self._tools = {}
        self._mimetypes = SubStorage(self._tools)
        self._languages = SubStorage(self._tools)
        self._content_ctags = SubStorage(self._tools)
        self._licenses = SubStorage(self._tools)
        self._content_metadata = SubStorage(self._tools)
        self._revision_intrinsic_metadata = SubStorage(self._tools)
        self._origin_intrinsic_metadata = SubStorage(self._tools)

    def check_config(self, *, check_write):
        return True

    def content_mimetype_missing(self, mimetypes):
        yield from self._mimetypes.missing(mimetypes)

    def content_mimetype_get_range(
            self, start, end, indexer_configuration_id, limit=1000):
        return self._mimetypes.get_range(
            start, end, indexer_configuration_id, limit)

    def content_mimetype_add(self, mimetypes, conflict_update=False):
        if not all(isinstance(x['id'], bytes) for x in mimetypes):
            raise TypeError('identifiers must be bytes.')
        self._mimetypes.add(mimetypes, conflict_update)

    def content_mimetype_get(self, ids):
        yield from self._mimetypes.get(ids)

    def content_language_missing(self, languages):
        yield from self._languages.missing(languages)

    def content_language_get(self, ids):
        yield from self._languages.get(ids)

    def content_language_add(self, languages, conflict_update=False):
        if not all(isinstance(x['id'], bytes) for x in languages):
            raise TypeError('identifiers must be bytes.')
        self._languages.add(languages, conflict_update)

    def content_ctags_missing(self, ctags):
        yield from self._content_ctags.missing(ctags)

    def content_ctags_get(self, ids):
        for item in self._content_ctags.get(ids):
            for item_ctags_item in item['ctags']:
                yield {
                    'id': item['id'],
                    'tool': item['tool'],
                    **item_ctags_item
                }

    def content_ctags_add(self, ctags, conflict_update=False):
        if not all(isinstance(x['id'], bytes) for x in ctags):
            raise TypeError('identifiers must be bytes.')
        self._content_ctags.add_merge(ctags, conflict_update, 'ctags')

    def content_ctags_search(self, expression,
                             limit=10, last_sha1=None):
        nb_matches = 0
        for ((id_, tool_id), item) in \
                sorted(self._content_ctags._data.items()):
            if id_ <= (last_sha1 or bytes(0 for _ in range(SHA1_DIGEST_SIZE))):
                continue
            for ctags_item in item['ctags']:
                if ctags_item['name'] != expression:
                    continue
                nb_matches += 1
                yield {
                    'id': id_,
                    'tool': _transform_tool(self._tools[tool_id]),
                    **ctags_item
                }
                if nb_matches >= limit:
                    return

    def content_fossology_license_get(self, ids):
        # Rewrites the output of SubStorage.get from the old format to
        # the new one. SubStorage.get should be updated once all other
        # *_get methods use the new format.
        # See: https://forge.softwareheritage.org/T1433
        res = {}
        for d in self._licenses.get(ids):
            res.setdefault(d.pop('id'), []).append(d)
        for (id_, facts) in res.items():
            yield {id_: facts}

    def content_fossology_license_add(self, licenses, conflict_update=False):
        if not all(isinstance(x['id'], bytes) for x in licenses):
            raise TypeError('identifiers must be bytes.')
        self._licenses.add_merge(licenses, conflict_update, 'licenses')

    def content_fossology_license_get_range(
            self, start, end, indexer_configuration_id, limit=1000):
        return self._licenses.get_range(
            start, end, indexer_configuration_id, limit)

    def content_metadata_missing(self, metadata):
        yield from self._content_metadata.missing(metadata)

    def content_metadata_get(self, ids):
        yield from self._content_metadata.get(ids)

    def content_metadata_add(self, metadata, conflict_update=False):
        if not all(isinstance(x['id'], bytes) for x in metadata):
            raise TypeError('identifiers must be bytes.')
        self._content_metadata.add(metadata, conflict_update)

    def revision_intrinsic_metadata_missing(self, metadata):
        yield from self._revision_intrinsic_metadata.missing(metadata)

    def revision_intrinsic_metadata_get(self, ids):
        yield from self._revision_intrinsic_metadata.get(ids)

    def revision_intrinsic_metadata_add(self, metadata, conflict_update=False):
        if not all(isinstance(x['id'], bytes) for x in metadata):
            raise TypeError('identifiers must be bytes.')
        self._revision_intrinsic_metadata.add(metadata, conflict_update)

    def revision_intrinsic_metadata_delete(self, entries):
        self._revision_intrinsic_metadata.delete(entries)

    def origin_intrinsic_metadata_get(self, ids):
        yield from self._origin_intrinsic_metadata.get(ids)

    def origin_intrinsic_metadata_add(self, metadata,
                                      conflict_update=False):
        self._origin_intrinsic_metadata.add(metadata, conflict_update)

    def origin_intrinsic_metadata_delete(self, entries):
        self._origin_intrinsic_metadata.delete(entries)

    def origin_intrinsic_metadata_search_fulltext(
            self, conjunction, limit=100):
        # A very crude fulltext search implementation, but that's enough
        # to work on English metadata
        tokens_re = re.compile('[a-zA-Z0-9]+')
        search_tokens = list(itertools.chain(
            *map(tokens_re.findall, conjunction)))

        def rank(data):
            # Tokenize the metadata
            text = json.dumps(data['metadata'])
            text_tokens = tokens_re.findall(text)
            text_token_occurences = Counter(text_tokens)

            # Count the number of occurrences of search tokens in the text
            score = 0
            for search_token in search_tokens:
                if text_token_occurences[search_token] == 0:
                    # Search token is not in the text.
                    return 0
                score += text_token_occurences[search_token]

            # Normalize according to the text's length
            return score / math.log(len(text_tokens))

        results = [(rank(data), data)
                   for data in self._origin_intrinsic_metadata.get_all()]
        results = [(rank_, data) for (rank_, data) in results if rank_ > 0]
        results.sort(key=operator.itemgetter(0),  # Don't try to order 'data'
                     reverse=True)
        for (rank_, result) in results[:limit]:
            yield result

    def origin_intrinsic_metadata_search_by_producer(
            self, page_token='', limit=100, ids_only=False,
            mappings=None, tool_ids=None):
        assert isinstance(page_token, str)
        nb_results = 0
        if mappings is not None:
            mappings = frozenset(mappings)
        if tool_ids is not None:
            tool_ids = frozenset(tool_ids)
        origins = []

        # we go to limit+1 to check whether we should add next_page_token in
        # the response
        for entry in self._origin_intrinsic_metadata.get_all():
            if entry['id'] <= page_token:
                continue
            if nb_results >= (limit + 1):
                break
            if mappings is not None and mappings.isdisjoint(entry['mappings']):
                continue
            if tool_ids is not None and entry['tool']['id'] not in tool_ids:
                continue
            origins.append(entry)
            nb_results += 1

        result = {}
        if len(origins) > limit:
            origins = origins[:limit]
            result['next_page_token'] = origins[-1]['id']
        if ids_only:
            origins = [origin['id'] for origin in origins]
        result['origins'] = origins
        return result

    def origin_intrinsic_metadata_stats(self):
        mapping_count = {m: 0 for m in MAPPING_NAMES}
        total = non_empty = 0
        for data in self._origin_intrinsic_metadata.get_all():
            total += 1
            if set(data['metadata']) - {'@context'}:
                non_empty += 1
            for mapping in data['mappings']:
                mapping_count[mapping] += 1
        return {
            'per_mapping': mapping_count,
            'total': total,
            'non_empty': non_empty
        }

    def indexer_configuration_add(self, tools):
        inserted = []
        for tool in tools:
            tool = tool.copy()
            id_ = self._tool_key(tool)
            tool['id'] = id_
            self._tools[id_] = tool
            inserted.append(tool)
        return inserted

    def indexer_configuration_get(self, tool):
        return self._tools.get(self._tool_key(tool))

    def _tool_key(self, tool):
        return hash((tool['tool_name'], tool['tool_version'],
                     json.dumps(tool['tool_configuration'], sort_keys=True)))



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from unittest.mock import patch

from swh.model.hashutil import hash_to_bytes

from swh.indexer.metadata import OriginMetadataIndexer

from .utils import YARN_PARSER_METADATA
from .test_metadata import REVISION_METADATA_CONFIG


def test_origin_metadata_indexer(idx_storage, storage, obj_storage):

    indexer = OriginMetadataIndexer(config=REVISION_METADATA_CONFIG)
    indexer.run(["https://github.com/librariesio/yarn-parser"])

    origin = 'https://github.com/librariesio/yarn-parser'
    rev_id = hash_to_bytes('8dbb6aeb036e7fd80664eb8bfd1507881af1ba9f')

    rev_metadata = {
        'id': rev_id,
        'metadata': YARN_PARSER_METADATA,
        'mappings': ['npm'],
    }
    origin_metadata = {
        'id': origin,
        'from_revision': rev_id,
        'metadata': YARN_PARSER_METADATA,
        'mappings': ['npm'],
    }

    results = list(
        indexer.idx_storage.revision_intrinsic_metadata_get([rev_id]))
    for result in results:
        del result['tool']
    assert results == [rev_metadata]

    results = list(indexer.idx_storage.origin_intrinsic_metadata_get([
        origin]))
    for result in results:
        del result['tool']
    assert results == [origin_metadata]


def test_origin_metadata_indexer_duplicate_origin(
        idx_storage, storage, obj_storage):
    indexer = OriginMetadataIndexer(config=REVISION_METADATA_CONFIG)
    indexer.storage = storage
    indexer.idx_storage = idx_storage
    indexer.run(["https://github.com/librariesio/yarn-parser"])

    indexer.run(["https://github.com/librariesio/yarn-parser"]*2)

    origin = 'https://github.com/librariesio/yarn-parser'
    rev_id = hash_to_bytes('8dbb6aeb036e7fd80664eb8bfd1507881af1ba9f')

    results = list(
        indexer.idx_storage.revision_intrinsic_metadata_get([rev_id]))
    assert len(results) == 1

    results = list(indexer.idx_storage.origin_intrinsic_metadata_get([
        origin]))
    assert len(results) == 1


def test_origin_metadata_indexer_missing_head(
        idx_storage, storage, obj_storage):

    storage.origin_add([{
        'type': 'git',
        'url': 'https://example.com'
    }])

    indexer = OriginMetadataIndexer(config=REVISION_METADATA_CONFIG)
    indexer.run(["https://example.com"])

    origin = 'https://example.com'

    results = list(indexer.idx_storage.origin_intrinsic_metadata_get([
        origin]))
    assert results == []


def test_origin_metadata_indexer_partial_missing_head(
        idx_storage, storage, obj_storage):

    storage.origin_add([{
        'type': 'git',
        'url': 'https://example.com'
    }])

    indexer = OriginMetadataIndexer(config=REVISION_METADATA_CONFIG)
    indexer.run(["https://example.com",
                 "https://github.com/librariesio/yarn-parser"])

    origin1 = 'https://example.com'
    origin2 = 'https://github.com/librariesio/yarn-parser'
    rev_id = hash_to_bytes('8dbb6aeb036e7fd80664eb8bfd1507881af1ba9f')

    rev_metadata = {
        'id': rev_id,
        'metadata': YARN_PARSER_METADATA,
        'mappings': ['npm'],
    }
    origin_metadata = {
        'id': origin2,
        'from_revision': rev_id,
        'metadata': YARN_PARSER_METADATA,
        'mappings': ['npm'],
    }

    results = list(
        indexer.idx_storage.revision_intrinsic_metadata_get([rev_id]))
    for result in results:
        del result['tool']
    assert results == [rev_metadata]

    results = list(indexer.idx_storage.origin_intrinsic_metadata_get([
        origin1, origin2]))
    for result in results:
        del result['tool']
    assert results == [origin_metadata]


def test_origin_metadata_indexer_duplicate_revision(
        idx_storage, storage, obj_storage):
    indexer = OriginMetadataIndexer(config=REVISION_METADATA_CONFIG)
    indexer.storage = storage
    indexer.idx_storage = idx_storage
    indexer.run(["https://github.com/librariesio/yarn-parser",
                 "https://github.com/librariesio/yarn-parser.git"])

    origin1 = 'https://github.com/librariesio/yarn-parser'
    origin2 = 'https://github.com/librariesio/yarn-parser.git'
    rev_id = hash_to_bytes('8dbb6aeb036e7fd80664eb8bfd1507881af1ba9f')

    results = list(
        indexer.idx_storage.revision_intrinsic_metadata_get([rev_id]))
    assert len(results) == 1

    results = list(indexer.idx_storage.origin_intrinsic_metadata_get(
        [origin1, origin2]))
    assert len(results) == 2


def test_origin_metadata_indexer_no_metadata_file(
        idx_storage, storage, obj_storage):

    indexer = OriginMetadataIndexer(config=REVISION_METADATA_CONFIG)
    with patch('swh.indexer.metadata_dictionary.npm.NpmMapping.filename',
               b'foo.json'):
        indexer.run(["https://github.com/librariesio/yarn-parser"])

    origin = 'https://github.com/librariesio/yarn-parser'
    rev_id = hash_to_bytes('8dbb6aeb036e7fd80664eb8bfd1507881af1ba9f')

    results = list(
        indexer.idx_storage.revision_intrinsic_metadata_get([rev_id]))
    assert results == []

    results = list(indexer.idx_storage.origin_intrinsic_metadata_get([
        origin]))
    assert results == []


def test_origin_metadata_indexer_no_metadata(
        idx_storage, storage, obj_storage):

    indexer = OriginMetadataIndexer(config=REVISION_METADATA_CONFIG)
    with patch('swh.indexer.metadata.RevisionMetadataIndexer'
               '.translate_revision_intrinsic_metadata',
               return_value=(['npm'], {'@context': 'foo'})):
        indexer.run(["https://github.com/librariesio/yarn-parser"])

    origin = 'https://github.com/librariesio/yarn-parser'
    rev_id = hash_to_bytes('8dbb6aeb036e7fd80664eb8bfd1507881af1ba9f')

    results = list(
        indexer.idx_storage.revision_intrinsic_metadata_get([rev_id]))
    assert results == []

    results = list(indexer.idx_storage.origin_intrinsic_metadata_get([
        origin]))
    assert results == []


def test_origin_metadata_indexer_error(
        idx_storage, storage, obj_storage):

    indexer = OriginMetadataIndexer(config=REVISION_METADATA_CONFIG)
    with patch('swh.indexer.metadata.RevisionMetadataIndexer'
               '.translate_revision_intrinsic_metadata',
               return_value=None):
        indexer.run(["https://github.com/librariesio/yarn-parser"])

    origin = 'https://github.com/librariesio/yarn-parser'
    rev_id = hash_to_bytes('8dbb6aeb036e7fd80664eb8bfd1507881af1ba9f')

    results = list(
        indexer.idx_storage.revision_intrinsic_metadata_get([rev_id]))
    assert results == []

    results = list(indexer.idx_storage.origin_intrinsic_metadata_get([
        origin]))
    assert results == []


def test_origin_metadata_indexer_delete_metadata(
        idx_storage, storage, obj_storage):

    indexer = OriginMetadataIndexer(config=REVISION_METADATA_CONFIG)
    indexer.run(["https://github.com/librariesio/yarn-parser"])

    origin = 'https://github.com/librariesio/yarn-parser'
    rev_id = hash_to_bytes('8dbb6aeb036e7fd80664eb8bfd1507881af1ba9f')

    results = list(
        indexer.idx_storage.revision_intrinsic_metadata_get([rev_id]))
    assert results != []

    results = list(indexer.idx_storage.origin_intrinsic_metadata_get([
        origin]))
    assert results != []

    with patch('swh.indexer.metadata_dictionary.npm.NpmMapping.filename',
               b'foo.json'):
        indexer.run(["https://github.com/librariesio/yarn-parser"])

    results = list(
        indexer.idx_storage.revision_intrinsic_metadata_get([rev_id]))
    assert results == []

    results = list(indexer.idx_storage.origin_intrinsic_metadata_get([
        origin]))
    assert results == []



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import collections
import csv
import itertools
import json
import os.path
import re

import swh.indexer
from pyld import jsonld

_DATA_DIR = os.path.join(os.path.dirname(swh.indexer.__file__), 'data')

CROSSWALK_TABLE_PATH = os.path.join(_DATA_DIR, 'codemeta', 'crosswalk.csv')

CODEMETA_CONTEXT_PATH = os.path.join(_DATA_DIR, 'codemeta', 'codemeta.jsonld')


with open(CODEMETA_CONTEXT_PATH) as fd:
    CODEMETA_CONTEXT = json.load(fd)

CODEMETA_CONTEXT_URL = 'https://doi.org/10.5063/schema/codemeta-2.0'
CODEMETA_ALTERNATE_CONTEXT_URLS = {
    ('https://raw.githubusercontent.com/codemeta/codemeta/'
     'master/codemeta.jsonld')
}
CODEMETA_URI = 'https://codemeta.github.io/terms/'
SCHEMA_URI = 'http://schema.org/'


PROPERTY_BLACKLIST = {
    # CodeMeta properties that we cannot properly represent.
    SCHEMA_URI + 'softwareRequirements',
    CODEMETA_URI + 'softwareSuggestions',

    # Duplicate of 'author'
    SCHEMA_URI + 'creator',
    }

_codemeta_field_separator = re.compile(r'\s*[,/]\s*')


def make_absolute_uri(local_name):
    definition = CODEMETA_CONTEXT['@context'][local_name]
    if isinstance(definition, str):
        return definition
    elif isinstance(definition, dict):
        prefixed_name = definition['@id']
        (prefix, local_name) = prefixed_name.split(':')
        if prefix == 'schema':
            canonical_name = SCHEMA_URI + local_name
        elif prefix == 'codemeta':
            canonical_name = CODEMETA_URI + local_name
        else:
            assert False, prefix
        return canonical_name
    else:
        assert False, definition


def _read_crosstable(fd):
    reader = csv.reader(fd)
    try:
        header = next(reader)
    except StopIteration:
        raise ValueError('empty file')

    data_sources = set(header) - {'Parent Type', 'Property',
                                  'Type', 'Description'}
    assert 'codemeta-V1' in data_sources

    codemeta_translation = {data_source: {} for data_source in data_sources}
    terms = set()

    for line in reader:  # For each canonical name
        local_name = dict(zip(header, line))['Property']
        if not local_name:
            continue
        canonical_name = make_absolute_uri(local_name)
        if canonical_name in PROPERTY_BLACKLIST:
            continue
        terms.add(canonical_name)
        for (col, value) in zip(header, line):  # For each cell in the row
            if col in data_sources:
                # If that's not the parentType/property/type/description
                for local_name in _codemeta_field_separator.split(value):
                    # For each of the data source's properties that maps
                    # to this canonical name
                    if local_name.strip():
                        codemeta_translation[col][local_name.strip()] = \
                                canonical_name

    return (terms, codemeta_translation)


with open(CROSSWALK_TABLE_PATH) as fd:
    (CODEMETA_TERMS, CROSSWALK_TABLE) = _read_crosstable(fd)


def _document_loader(url):
    """Document loader for pyld.

    Reads the local codemeta.jsonld file instead of fetching it
    from the Internet every single time."""
    if url == CODEMETA_CONTEXT_URL or url in CODEMETA_ALTERNATE_CONTEXT_URLS:
        return {
                'contextUrl': None,
                'documentUrl': url,
                'document': CODEMETA_CONTEXT,
                }
    elif url == CODEMETA_URI:
        raise Exception('{} is CodeMeta\'s URI, use {} as context url'.format(
            CODEMETA_URI, CODEMETA_CONTEXT_URL))
    else:
        raise Exception(url)


def compact(doc):
    """Same as `pyld.jsonld.compact`, but in the context of CodeMeta."""
    return jsonld.compact(doc, CODEMETA_CONTEXT_URL,
                          options={'documentLoader': _document_loader})


def expand(doc):
    """Same as `pyld.jsonld.expand`, but in the context of CodeMeta."""
    return jsonld.expand(doc,
                         options={'documentLoader': _document_loader})


def merge_documents(documents):
    """Takes a list of metadata dicts, each generated from a different
    metadata file, and merges them.

    Removes duplicates, if any."""
    documents = list(itertools.chain.from_iterable(map(expand, documents)))
    merged_document = collections.defaultdict(list)
    for document in documents:
        for (key, values) in document.items():
            if key == '@id':
                # @id does not get expanded to a list
                value = values

                # Only one @id is allowed, move it to sameAs
                if '@id' not in merged_document:
                    merged_document['@id'] = value
                elif value != merged_document['@id']:
                    if value not in merged_document[SCHEMA_URI + 'sameAs']:
                        merged_document[SCHEMA_URI + 'sameAs'].append(value)
            else:
                for value in values:
                    if value not in merged_document[key]:
                        merged_document[key].append(value)

    return compact(merged_document)



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
import unittest

import attr

from hypothesis import given, strategies, settings, HealthCheck

from swh.model.hashutil import hash_to_bytes

from swh.indexer.codemeta import CODEMETA_TERMS, CROSSWALK_TABLE
from swh.indexer.codemeta import merge_documents
from swh.indexer.metadata_dictionary import MAPPINGS
from swh.indexer.metadata_dictionary.base import merge_values
from swh.indexer.metadata_dictionary.maven import MavenMapping
from swh.indexer.metadata_dictionary.npm import NpmMapping
from swh.indexer.metadata_dictionary.ruby import GemspecMapping
from swh.indexer.metadata_detector import (
    detect_metadata
)
from swh.indexer.metadata import (
    ContentMetadataIndexer, RevisionMetadataIndexer
)

from .utils import (
    BASE_TEST_CONFIG, fill_obj_storage, fill_storage,
    YARN_PARSER_METADATA, json_document_strategy,
    xml_document_strategy,
)


TRANSLATOR_TOOL = {
    'name': 'swh-metadata-translator',
    'version': '0.0.2',
    'configuration': {
        'type': 'local',
        'context': 'NpmMapping'
    }
}


class ContentMetadataTestIndexer(ContentMetadataIndexer):
    """Specific Metadata whose configuration is enough to satisfy the
       indexing tests.
    """
    def parse_config_file(self, *args, **kwargs):
        assert False, 'should not be called; the rev indexer configures it.'


REVISION_METADATA_CONFIG = {
    **BASE_TEST_CONFIG,
    'tools': TRANSLATOR_TOOL,
}


class Metadata(unittest.TestCase):
    """
    Tests metadata_mock_tool tool for Metadata detection
    """
    def setUp(self):
        """
        shows the entire diff in the results
        """
        self.maxDiff = None
        self.npm_mapping = MAPPINGS['NpmMapping']()
        self.codemeta_mapping = MAPPINGS['CodemetaMapping']()
        self.maven_mapping = MAPPINGS['MavenMapping']()
        self.pkginfo_mapping = MAPPINGS['PythonPkginfoMapping']()
        self.gemspec_mapping = MAPPINGS['GemspecMapping']()

    def test_crosstable(self):
        self.assertEqual(CROSSWALK_TABLE['NodeJS'], {
            'repository': 'http://schema.org/codeRepository',
            'os': 'http://schema.org/operatingSystem',
            'cpu': 'http://schema.org/processorRequirements',
            'engines':
                'http://schema.org/processorRequirements',
            'author': 'http://schema.org/author',
            'author.email': 'http://schema.org/email',
            'author.name': 'http://schema.org/name',
            'contributor': 'http://schema.org/contributor',
            'keywords': 'http://schema.org/keywords',
            'license': 'http://schema.org/license',
            'version': 'http://schema.org/version',
            'description': 'http://schema.org/description',
            'name': 'http://schema.org/name',
            'bugs': 'https://codemeta.github.io/terms/issueTracker',
            'homepage': 'http://schema.org/url'
        })

    def test_merge_values(self):
        self.assertEqual(
            merge_values('a', 'b'),
            ['a', 'b'])
        self.assertEqual(
            merge_values(['a', 'b'], 'c'),
            ['a', 'b', 'c'])
        self.assertEqual(
            merge_values('a', ['b', 'c']),
            ['a', 'b', 'c'])

        self.assertEqual(
            merge_values({'@list': ['a']}, {'@list': ['b']}),
            {'@list': ['a', 'b']})
        self.assertEqual(
            merge_values({'@list': ['a', 'b']}, {'@list': ['c']}),
            {'@list': ['a', 'b', 'c']})

        with self.assertRaises(ValueError):
            merge_values({'@list': ['a']}, 'b')
        with self.assertRaises(ValueError):
            merge_values('a', {'@list': ['b']})
        with self.assertRaises(ValueError):
            merge_values({'@list': ['a']}, ['b'])
        with self.assertRaises(ValueError):
            merge_values(['a'], {'@list': ['b']})

        self.assertEqual(
            merge_values('a', None),
            'a')
        self.assertEqual(
            merge_values(['a', 'b'], None),
            ['a', 'b'])
        self.assertEqual(
            merge_values(None, ['b', 'c']),
            ['b', 'c'])
        self.assertEqual(
            merge_values({'@list': ['a']}, None),
            {'@list': ['a']})
        self.assertEqual(
            merge_values(None, {'@list': ['a']}),
            {'@list': ['a']})

    def test_compute_metadata_none(self):
        """
        testing content empty content is empty
        should return None
        """
        # given
        content = b""

        # None if no metadata was found or an error occurred
        declared_metadata = None
        # when
        result = self.npm_mapping.translate(content)
        # then
        self.assertEqual(declared_metadata, result)

    def test_compute_metadata_npm(self):
        """
        testing only computation of metadata with hard_mapping_npm
        """
        # given
        content = b"""
            {
                "name": "test_metadata",
                "version": "0.0.2",
                "description": "Simple package.json test for indexer",
                  "repository": {
                    "type": "git",
                    "url": "https://github.com/moranegg/metadata_test"
                },
                "author": {
                    "email": "moranegg@example.com",
                    "name": "Morane G"
                }
            }
        """
        declared_metadata = {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
            'name': 'test_metadata',
            'version': '0.0.2',
            'description': 'Simple package.json test for indexer',
            'codeRepository':
                'git+https://github.com/moranegg/metadata_test',
            'author': [{
                'type': 'Person',
                'name': 'Morane G',
                'email': 'moranegg@example.com',
            }],
        }

        # when
        result = self.npm_mapping.translate(content)
        # then
        self.assertEqual(declared_metadata, result)

    def test_merge_documents(self):
        """
        Test the creation of a coherent minimal metadata set
        """
        # given
        metadata_list = [{
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'name': 'test_1',
            'version': '0.0.2',
            'description': 'Simple package.json test for indexer',
            'codeRepository':
                'git+https://github.com/moranegg/metadata_test',
        }, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'name': 'test_0_1',
            'version': '0.0.2',
            'description': 'Simple package.json test for indexer',
            'codeRepository':
                'git+https://github.com/moranegg/metadata_test'
        }, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'name': 'test_metadata',
            'version': '0.0.2',
            'author': {
                'type': 'Person',
                'name': 'moranegg',
            },
        }]

        # when
        results = merge_documents(metadata_list)

        # then
        expected_results = {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            "version": '0.0.2',
            "description": 'Simple package.json test for indexer',
            "name": ['test_1', 'test_0_1', 'test_metadata'],
            "author": [{
                'type': 'Person',
                'name': 'moranegg'
            }],
            "codeRepository":
                'git+https://github.com/moranegg/metadata_test',
        }
        self.assertEqual(expected_results, results)

    def test_merge_documents_ids(self):
        # given
        metadata_list = [{
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'id': 'http://example.org/test1',
            'name': 'test_1',
        }, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'id': 'http://example.org/test2',
            'name': 'test_2',
        }]

        # when
        results = merge_documents(metadata_list)

        # then
        expected_results = {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'id': 'http://example.org/test1',
            'schema:sameAs': 'http://example.org/test2',
            "name": ['test_1', 'test_2']
        }
        self.assertEqual(expected_results, results)

    def test_merge_documents_duplicate_ids(self):
        # given
        metadata_list = [{
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'id': 'http://example.org/test1',
            'name': 'test_1',
        }, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'id': 'http://example.org/test1',
            'name': 'test_1b',
        }, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'id': 'http://example.org/test2',
            'name': 'test_2',
        }]

        # when
        results = merge_documents(metadata_list)

        # then
        expected_results = {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'id': 'http://example.org/test1',
            'schema:sameAs': 'http://example.org/test2',
            "name": ['test_1', 'test_1b', 'test_2']
        }
        self.assertEqual(expected_results, results)

    def test_index_content_metadata_npm(self):
        """
        testing NPM with package.json
        - one sha1 uses a file that can't be translated to metadata and
          should return None in the translated metadata
        """
        # given
        sha1s = [
            hash_to_bytes('26a9f72a7c87cc9205725cfd879f514ff4f3d8d5'),
            hash_to_bytes('d4c647f0fc257591cc9ba1722484229780d1c607'),
            hash_to_bytes('02fb2c89e14f7fab46701478c83779c7beb7b069'),
        ]
        # this metadata indexer computes only metadata for package.json
        # in npm context with a hard mapping
        config = BASE_TEST_CONFIG.copy()
        config['tools'] = [TRANSLATOR_TOOL]
        metadata_indexer = ContentMetadataTestIndexer(config=config)
        fill_obj_storage(metadata_indexer.objstorage)
        fill_storage(metadata_indexer.storage)

        # when
        metadata_indexer.run(sha1s, policy_update='ignore-dups')
        results = list(metadata_indexer.idx_storage.content_metadata_get(
            sha1s))

        expected_results = [{
            'metadata': {
                '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
                'type': 'SoftwareSourceCode',
                'codeRepository':
                    'git+https://github.com/moranegg/metadata_test',
                'description': 'Simple package.json test for indexer',
                'name': 'test_metadata',
                'version': '0.0.1'
            },
            'id': hash_to_bytes('26a9f72a7c87cc9205725cfd879f514ff4f3d8d5'),
            }, {
            'metadata': {
                '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
                'type': 'SoftwareSourceCode',
                'issueTracker':
                    'https://github.com/npm/npm/issues',
                'author': [{
                    'type': 'Person',
                    'name': 'Isaac Z. Schlueter',
                    'email': 'i@izs.me',
                    'url': 'http://blog.izs.me',
                }],
                'codeRepository':
                    'git+https://github.com/npm/npm',
                'description': 'a package manager for JavaScript',
                'license': 'https://spdx.org/licenses/Artistic-2.0',
                'version': '5.0.3',
                'name': 'npm',
                'keywords': [
                    'install',
                    'modules',
                    'package manager',
                    'package.json'
                ],
                'url': 'https://docs.npmjs.com/'
            },
            'id': hash_to_bytes('d4c647f0fc257591cc9ba1722484229780d1c607')
        }]

        for result in results:
            del result['tool']

        # The assertion below returns False sometimes because of nested lists
        self.assertEqual(expected_results, results)

    def test_npm_bugs_normalization(self):
        # valid dictionary
        package_json = b"""{
            "name": "foo",
            "bugs": {
                "url": "https://github.com/owner/project/issues",
                "email": "foo@example.com"
            }
        }"""
        result = self.npm_mapping.translate(package_json)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'name': 'foo',
            'issueTracker': 'https://github.com/owner/project/issues',
            'type': 'SoftwareSourceCode',
        })

        # "invalid" dictionary
        package_json = b"""{
            "name": "foo",
            "bugs": {
                "email": "foo@example.com"
            }
        }"""
        result = self.npm_mapping.translate(package_json)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'name': 'foo',
            'type': 'SoftwareSourceCode',
        })

        # string
        package_json = b"""{
            "name": "foo",
            "bugs": "https://github.com/owner/project/issues"
        }"""
        result = self.npm_mapping.translate(package_json)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'name': 'foo',
            'issueTracker': 'https://github.com/owner/project/issues',
            'type': 'SoftwareSourceCode',
        })

    def test_npm_repository_normalization(self):
        # normal
        package_json = b"""{
            "name": "foo",
            "repository": {
                "type" : "git",
                "url" : "https://github.com/npm/cli.git"
            }
        }"""
        result = self.npm_mapping.translate(package_json)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'name': 'foo',
            'codeRepository': 'git+https://github.com/npm/cli.git',
            'type': 'SoftwareSourceCode',
        })

        # missing url
        package_json = b"""{
            "name": "foo",
            "repository": {
                "type" : "git"
            }
        }"""
        result = self.npm_mapping.translate(package_json)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'name': 'foo',
            'type': 'SoftwareSourceCode',
        })

        # github shortcut
        package_json = b"""{
            "name": "foo",
            "repository": "github:npm/cli"
        }"""
        result = self.npm_mapping.translate(package_json)
        expected_result = {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'name': 'foo',
            'codeRepository': 'git+https://github.com/npm/cli.git',
            'type': 'SoftwareSourceCode',
        }
        self.assertEqual(result, expected_result)

        # github shortshortcut
        package_json = b"""{
            "name": "foo",
            "repository": "npm/cli"
        }"""
        result = self.npm_mapping.translate(package_json)
        self.assertEqual(result, expected_result)

        # gitlab shortcut
        package_json = b"""{
            "name": "foo",
            "repository": "gitlab:user/repo"
        }"""
        result = self.npm_mapping.translate(package_json)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'name': 'foo',
            'codeRepository': 'git+https://gitlab.com/user/repo.git',
            'type': 'SoftwareSourceCode',
        })

    def test_detect_metadata_package_json(self):
        # given
        df = [{
                'sha1_git': b'abc',
                'name': b'index.js',
                'target': b'abc',
                'length': 897,
                'status': 'visible',
                'type': 'file',
                'perms': 33188,
                'dir_id': b'dir_a',
                'sha1': b'bcd'
            },
            {
                'sha1_git': b'aab',
                'name': b'package.json',
                'target': b'aab',
                'length': 712,
                'status': 'visible',
                'type': 'file',
                'perms': 33188,
                'dir_id': b'dir_a',
                'sha1': b'cde'
        }]
        # when
        results = detect_metadata(df)

        expected_results = {
            'NpmMapping': [
                b'cde'
            ]
        }
        # then
        self.assertEqual(expected_results, results)

    def test_compute_metadata_valid_codemeta(self):
        raw_content = (
            b"""{
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "@type": "SoftwareSourceCode",
            "identifier": "CodeMeta",
            "description": "CodeMeta is a concept vocabulary that can be used to standardize the exchange of software metadata across repositories and organizations.",
            "name": "CodeMeta: Minimal metadata schemas for science software and code, in JSON-LD",
            "codeRepository": "https://github.com/codemeta/codemeta",
            "issueTracker": "https://github.com/codemeta/codemeta/issues",
            "license": "https://spdx.org/licenses/Apache-2.0",
            "version": "2.0",
            "author": [
              {
                "@type": "Person",
                "givenName": "Carl",
                "familyName": "Boettiger",
                "email": "cboettig@gmail.com",
                "@id": "http://orcid.org/0000-0002-1642-628X"
              },
              {
                "@type": "Person",
                "givenName": "Matthew B.",
                "familyName": "Jones",
                "email": "jones@nceas.ucsb.edu",
                "@id": "http://orcid.org/0000-0003-0077-4738"
              }
            ],
            "maintainer": {
              "@type": "Person",
              "givenName": "Carl",
              "familyName": "Boettiger",
              "email": "cboettig@gmail.com",
              "@id": "http://orcid.org/0000-0002-1642-628X"
            },
            "contIntegration": "https://travis-ci.org/codemeta/codemeta",
            "developmentStatus": "active",
            "downloadUrl": "https://github.com/codemeta/codemeta/archive/2.0.zip",
            "funder": {
                "@id": "https://doi.org/10.13039/100000001",
                "@type": "Organization",
                "name": "National Science Foundation"
            },
            "funding":"1549758; Codemeta: A Rosetta Stone for Metadata in Scientific Software",
            "keywords": [
              "metadata",
              "software"
            ],
            "version":"2.0",
            "dateCreated":"2017-06-05",
            "datePublished":"2017-06-05",
            "programmingLanguage": "JSON-LD"
          }""") # noqa
        expected_result = {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "type": "SoftwareSourceCode",
            "identifier": "CodeMeta",
            "description":
                "CodeMeta is a concept vocabulary that can "
                "be used to standardize the exchange of software metadata "
                "across repositories and organizations.",
            "name":
                "CodeMeta: Minimal metadata schemas for science "
                "software and code, in JSON-LD",
            "codeRepository": "https://github.com/codemeta/codemeta",
            "issueTracker": "https://github.com/codemeta/codemeta/issues",
            "license": "https://spdx.org/licenses/Apache-2.0",
            "version": "2.0",
            "author": [
              {
                "type": "Person",
                "givenName": "Carl",
                "familyName": "Boettiger",
                "email": "cboettig@gmail.com",
                "id": "http://orcid.org/0000-0002-1642-628X"
              },
              {
                "type": "Person",
                "givenName": "Matthew B.",
                "familyName": "Jones",
                "email": "jones@nceas.ucsb.edu",
                "id": "http://orcid.org/0000-0003-0077-4738"
              }
            ],
            "maintainer": {
              "type": "Person",
              "givenName": "Carl",
              "familyName": "Boettiger",
              "email": "cboettig@gmail.com",
              "id": "http://orcid.org/0000-0002-1642-628X"
            },
            "contIntegration": "https://travis-ci.org/codemeta/codemeta",
            "developmentStatus": "active",
            "downloadUrl":
                "https://github.com/codemeta/codemeta/archive/2.0.zip",
            "funder": {
                "id": "https://doi.org/10.13039/100000001",
                "type": "Organization",
                "name": "National Science Foundation"
            },
            "funding": "1549758; Codemeta: A Rosetta Stone for Metadata "
                "in Scientific Software",
            "keywords": [
              "metadata",
              "software"
            ],
            "version": "2.0",
            "dateCreated": "2017-06-05",
            "datePublished": "2017-06-05",
            "programmingLanguage": "JSON-LD"
          }
        result = self.codemeta_mapping.translate(raw_content)
        self.assertEqual(result, expected_result)

    def test_compute_metadata_codemeta_alternate_context(self):
        raw_content = (
            b"""{
            "@context": "https://raw.githubusercontent.com/codemeta/codemeta/master/codemeta.jsonld",
            "@type": "SoftwareSourceCode",
            "identifier": "CodeMeta"
        }""")  # noqa
        expected_result = {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "type": "SoftwareSourceCode",
            "identifier": "CodeMeta",
        }
        result = self.codemeta_mapping.translate(raw_content)
        self.assertEqual(result, expected_result)

    def test_compute_metadata_maven(self):
        raw_content = b"""
        <project>
          <name>Maven Default Project</name>
          <modelVersion>4.0.0</modelVersion>
          <groupId>com.mycompany.app</groupId>
          <artifactId>my-app</artifactId>
          <version>1.2.3</version>
          <repositories>
            <repository>
              <id>central</id>
              <name>Maven Repository Switchboard</name>
              <layout>default</layout>
              <url>http://repo1.maven.org/maven2</url>
              <snapshots>
                <enabled>false</enabled>
              </snapshots>
            </repository>
          </repositories>
          <licenses>
            <license>
              <name>Apache License, Version 2.0</name>
              <url>https://www.apache.org/licenses/LICENSE-2.0.txt</url>
              <distribution>repo</distribution>
              <comments>A business-friendly OSS license</comments>
            </license>
          </licenses>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
            'name': 'Maven Default Project',
            'identifier': 'com.mycompany.app',
            'version': '1.2.3',
            'license': 'https://www.apache.org/licenses/LICENSE-2.0.txt',
            'codeRepository':
                'http://repo1.maven.org/maven2/com/mycompany/app/my-app',
        })

    def test_compute_metadata_maven_empty(self):
        raw_content = b"""
        <project>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
        })

    def test_compute_metadata_maven_almost_empty(self):
        raw_content = b"""
        <project>
          <foo/>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
        })

    def test_compute_metadata_maven_invalid_xml(self):
        expected_warning = (
            'WARNING:swh.indexer.metadata_dictionary.maven.MavenMapping:'
            'Error parsing XML from foo')

        raw_content = b"""
        <project>"""
        with self.assertLogs('swh.indexer.metadata_dictionary',
                             level='WARNING') as cm:
            result = MAPPINGS["MavenMapping"]('foo').translate(raw_content)
            self.assertEqual(cm.output, [expected_warning])
        self.assertEqual(result, None)

        raw_content = b"""
        """
        with self.assertLogs('swh.indexer.metadata_dictionary',
                             level='WARNING') as cm:
            result = MAPPINGS["MavenMapping"]('foo').translate(raw_content)
            self.assertEqual(cm.output, [expected_warning])
        self.assertEqual(result, None)

    def test_compute_metadata_maven_unknown_encoding(self):
        expected_warning = (
            'WARNING:swh.indexer.metadata_dictionary.maven.MavenMapping:'
            'Error detecting XML encoding from foo')

        raw_content = b"""<?xml version="1.0" encoding="foo"?>
        <project>
        </project>"""
        with self.assertLogs('swh.indexer.metadata_dictionary',
                             level='WARNING') as cm:
            result = MAPPINGS["MavenMapping"]('foo').translate(raw_content)
            self.assertEqual(cm.output, [expected_warning])
        self.assertEqual(result, None)

        raw_content = b"""<?xml version="1.0" encoding="UTF-7"?>
        <project>
        </project>"""
        with self.assertLogs('swh.indexer.metadata_dictionary',
                             level='WARNING') as cm:
            result = MAPPINGS["MavenMapping"]('foo').translate(raw_content)
            self.assertEqual(cm.output, [expected_warning])
        self.assertEqual(result, None)

    def test_compute_metadata_maven_invalid_encoding(self):
        expected_warning = (
            'WARNING:swh.indexer.metadata_dictionary.maven.MavenMapping:'
            'Error unidecoding XML from foo')

        raw_content = b"""<?xml version="1.0" encoding="UTF-8"?>
        <foo\xe5ct>
        </foo>"""
        with self.assertLogs('swh.indexer.metadata_dictionary',
                             level='WARNING') as cm:
            result = MAPPINGS["MavenMapping"]('foo').translate(raw_content)
            self.assertEqual(cm.output, [expected_warning])
        self.assertEqual(result, None)

    def test_compute_metadata_maven_minimal(self):
        raw_content = b"""
        <project>
          <name>Maven Default Project</name>
          <modelVersion>4.0.0</modelVersion>
          <groupId>com.mycompany.app</groupId>
          <artifactId>my-app</artifactId>
          <version>1.2.3</version>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
            'name': 'Maven Default Project',
            'identifier': 'com.mycompany.app',
            'version': '1.2.3',
            'codeRepository':
            'https://repo.maven.apache.org/maven2/com/mycompany/app/my-app',
        })

    def test_compute_metadata_maven_empty_nodes(self):
        raw_content = b"""
        <project>
          <name>Maven Default Project</name>
          <modelVersion>4.0.0</modelVersion>
          <groupId>com.mycompany.app</groupId>
          <artifactId>my-app</artifactId>
          <version>1.2.3</version>
          <repositories>
          </repositories>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
            'name': 'Maven Default Project',
            'identifier': 'com.mycompany.app',
            'version': '1.2.3',
            'codeRepository':
            'https://repo.maven.apache.org/maven2/com/mycompany/app/my-app',
        })

        raw_content = b"""
        <project>
          <name>Maven Default Project</name>
          <modelVersion>4.0.0</modelVersion>
          <groupId>com.mycompany.app</groupId>
          <artifactId>my-app</artifactId>
          <version></version>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
            'name': 'Maven Default Project',
            'identifier': 'com.mycompany.app',
            'codeRepository':
            'https://repo.maven.apache.org/maven2/com/mycompany/app/my-app',
        })

        raw_content = b"""
        <project>
          <name></name>
          <modelVersion>4.0.0</modelVersion>
          <groupId>com.mycompany.app</groupId>
          <artifactId>my-app</artifactId>
          <version>1.2.3</version>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
            'identifier': 'com.mycompany.app',
            'version': '1.2.3',
            'codeRepository':
            'https://repo.maven.apache.org/maven2/com/mycompany/app/my-app',
        })

        raw_content = b"""
        <project>
          <name>Maven Default Project</name>
          <modelVersion>4.0.0</modelVersion>
          <groupId>com.mycompany.app</groupId>
          <artifactId>my-app</artifactId>
          <version>1.2.3</version>
          <licenses>
          </licenses>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
            'name': 'Maven Default Project',
            'identifier': 'com.mycompany.app',
            'version': '1.2.3',
            'codeRepository':
            'https://repo.maven.apache.org/maven2/com/mycompany/app/my-app',
        })

        raw_content = b"""
        <project>
          <groupId></groupId>
          <version>1.2.3</version>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
            'version': '1.2.3',
        })

    def test_compute_metadata_maven_invalid_licenses(self):
        raw_content = b"""
        <project>
          <name>Maven Default Project</name>
          <modelVersion>4.0.0</modelVersion>
          <groupId>com.mycompany.app</groupId>
          <artifactId>my-app</artifactId>
          <version>1.2.3</version>
          <licenses>
            foo
          </licenses>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
            'name': 'Maven Default Project',
            'identifier': 'com.mycompany.app',
            'version': '1.2.3',
            'codeRepository':
            'https://repo.maven.apache.org/maven2/com/mycompany/app/my-app',
        })

    def test_compute_metadata_maven_multiple(self):
        '''Tests when there are multiple code repos and licenses.'''
        raw_content = b"""
        <project>
          <name>Maven Default Project</name>
          <modelVersion>4.0.0</modelVersion>
          <groupId>com.mycompany.app</groupId>
          <artifactId>my-app</artifactId>
          <version>1.2.3</version>
          <repositories>
            <repository>
              <id>central</id>
              <name>Maven Repository Switchboard</name>
              <layout>default</layout>
              <url>http://repo1.maven.org/maven2</url>
              <snapshots>
                <enabled>false</enabled>
              </snapshots>
            </repository>
            <repository>
              <id>example</id>
              <name>Example Maven Repo</name>
              <layout>default</layout>
              <url>http://example.org/maven2</url>
            </repository>
          </repositories>
          <licenses>
            <license>
              <name>Apache License, Version 2.0</name>
              <url>https://www.apache.org/licenses/LICENSE-2.0.txt</url>
              <distribution>repo</distribution>
              <comments>A business-friendly OSS license</comments>
            </license>
            <license>
              <name>MIT license</name>
              <url>https://opensource.org/licenses/MIT</url>
            </license>
          </licenses>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
            'name': 'Maven Default Project',
            'identifier': 'com.mycompany.app',
            'version': '1.2.3',
            'license': [
                'https://www.apache.org/licenses/LICENSE-2.0.txt',
                'https://opensource.org/licenses/MIT',
            ],
            'codeRepository': [
                'http://repo1.maven.org/maven2/com/mycompany/app/my-app',
                'http://example.org/maven2/com/mycompany/app/my-app',
            ]
        })

    def test_compute_metadata_pkginfo(self):
        raw_content = (b"""\
Metadata-Version: 2.1
Name: swh.core
Version: 0.0.49
Summary: Software Heritage core utilities
Home-page: https://forge.softwareheritage.org/diffusion/DCORE/
Author: Software Heritage developers
Author-email: swh-devel@inria.fr
License: UNKNOWN
Project-URL: Bug Reports, https://forge.softwareheritage.org/maniphest
Project-URL: Funding, https://www.softwareheritage.org/donate
Project-URL: Source, https://forge.softwareheritage.org/source/swh-core
Description: swh-core
        ========
       \x20
        core library for swh's modules:
        - config parser
        - hash computations
        - serialization
        - logging mechanism
       \x20
Platform: UNKNOWN
Classifier: Programming Language :: Python :: 3
Classifier: Intended Audience :: Developers
Classifier: License :: OSI Approved :: GNU General Public License v3 (GPLv3)
Classifier: Operating System :: OS Independent
Classifier: Development Status :: 5 - Production/Stable
Description-Content-Type: text/markdown
Provides-Extra: testing
""") # noqa
        result = self.pkginfo_mapping.translate(raw_content)
        self.assertCountEqual(result['description'], [
            'Software Heritage core utilities',  # note the comma here
            'swh-core\n'
            '========\n'
            '\n'
            "core library for swh's modules:\n"
            '- config parser\n'
            '- hash computations\n'
            '- serialization\n'
            '- logging mechanism\n'
            ''],
            result)
        del result['description']
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
            'url': 'https://forge.softwareheritage.org/diffusion/DCORE/',
            'name': 'swh.core',
            'author': [{
                'type': 'Person',
                'name': 'Software Heritage developers',
                'email': 'swh-devel@inria.fr',
            }],
            'version': '0.0.49',
        })

    def test_compute_metadata_pkginfo_utf8(self):
        raw_content = (b'''\
Metadata-Version: 1.1
Name: snowpyt
Description-Content-Type: UNKNOWN
Description: foo
        Hydrology N\xc2\xb083
''') # noqa
        result = self.pkginfo_mapping.translate(raw_content)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
            'name': 'snowpyt',
            'description': 'foo\nHydrology N°83',
        })

    def test_compute_metadata_pkginfo_keywords(self):
        raw_content = (b"""\
Metadata-Version: 2.1
Name: foo
Keywords: foo bar baz
""") # noqa
        result = self.pkginfo_mapping.translate(raw_content)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
            'name': 'foo',
            'keywords': ['foo', 'bar', 'baz'],
        })

    def test_compute_metadata_pkginfo_license(self):
        raw_content = (b"""\
Metadata-Version: 2.1
Name: foo
License: MIT
""") # noqa
        result = self.pkginfo_mapping.translate(raw_content)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
            'name': 'foo',
            'license': 'MIT',
        })

    def test_gemspec_base(self):
        raw_content = b"""
Gem::Specification.new do |s|
  s.name        = 'example'
  s.version     = '0.1.0'
  s.licenses    = ['MIT']
  s.summary     = "This is an example!"
  s.description = "Much longer explanation of the example!"
  s.authors     = ["Ruby Coder"]
  s.email       = 'rubycoder@example.com'
  s.files       = ["lib/example.rb"]
  s.homepage    = 'https://rubygems.org/gems/example'
  s.metadata    = { "source_code_uri" => "https://github.com/example/example" }
end"""
        result = self.gemspec_mapping.translate(raw_content)
        self.assertCountEqual(result.pop('description'), [
            "This is an example!",
            "Much longer explanation of the example!"
        ])
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
            'author': [
                {
                    'type': 'Person',
                    'name': 'Ruby Coder'
                }
            ],
            'name': 'example',
            'license': 'https://spdx.org/licenses/MIT',
            'codeRepository': 'https://rubygems.org/gems/example',
            'email': 'rubycoder@example.com',
            'version': '0.1.0',
        })

    def test_gemspec_two_author_fields(self):
        raw_content = b"""
Gem::Specification.new do |s|
  s.authors     = ["Ruby Coder1"]
  s.author      = "Ruby Coder2"
end"""
        result = self.gemspec_mapping.translate(raw_content)
        self.assertCountEqual(result.pop('author'), [
            {
                'type': 'Person',
                'name': 'Ruby Coder1'
            },
            {
                'type': 'Person',
                'name': 'Ruby Coder2'
            },
        ])
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
        })

    def test_gemspec_invalid_author(self):
        raw_content = b"""
Gem::Specification.new do |s|
  s.author      = ["Ruby Coder"]
end"""
        result = self.gemspec_mapping.translate(raw_content)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
        })
        raw_content = b"""
Gem::Specification.new do |s|
  s.author      = "Ruby Coder1",
end"""
        result = self.gemspec_mapping.translate(raw_content)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
        })
        raw_content = b"""
Gem::Specification.new do |s|
  s.authors     = ["Ruby Coder1", ["Ruby Coder2"]]
end"""
        result = self.gemspec_mapping.translate(raw_content)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
            'author': [
                {
                    'type': 'Person',
                    'name': 'Ruby Coder1'
                }
            ],
        })

    def test_gemspec_alternative_header(self):
        raw_content = b"""
require './lib/version'

Gem::Specification.new { |s|
  s.name = 'rb-system-with-aliases'
  s.summary = 'execute system commands with aliases'
}
"""
        result = self.gemspec_mapping.translate(raw_content)
        self.assertEqual(result, {
            '@context': 'https://doi.org/10.5063/schema/codemeta-2.0',
            'type': 'SoftwareSourceCode',
            'name': 'rb-system-with-aliases',
            'description': 'execute system commands with aliases',
        })

    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(json_document_strategy(keys=list(NpmMapping.mapping)))
    def test_npm_adversarial(self, doc):
        raw = json.dumps(doc).encode()
        self.npm_mapping.translate(raw)

    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(json_document_strategy(keys=CODEMETA_TERMS))
    def test_codemeta_adversarial(self, doc):
        raw = json.dumps(doc).encode()
        self.codemeta_mapping.translate(raw)

    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(xml_document_strategy(
        keys=list(MavenMapping.mapping),
        root='project',
        xmlns='http://maven.apache.org/POM/4.0.0'))
    def test_maven_adversarial(self, doc):
        self.maven_mapping.translate(doc)

    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(strategies.dictionaries(
        # keys
        strategies.one_of(
            strategies.text(),
            *map(strategies.just, GemspecMapping.mapping)
        ),
        # values
        strategies.recursive(
            strategies.characters(),
            lambda children: strategies.lists(children, 1)
        )
    ))
    def test_gemspec_adversarial(self, doc):
        parts = [b'Gem::Specification.new do |s|\n']
        for (k, v) in doc.items():
            parts.append('  s.{} = {}\n'.format(k, repr(v)).encode())
        parts.append(b'end\n')
        self.gemspec_mapping.translate(b''.join(parts))

    def test_revision_metadata_indexer(self):
        metadata_indexer = RevisionMetadataIndexer(
            config=REVISION_METADATA_CONFIG)
        fill_obj_storage(metadata_indexer.objstorage)
        fill_storage(metadata_indexer.storage)

        tool = metadata_indexer.idx_storage.indexer_configuration_get(
            {'tool_'+k: v for (k, v) in TRANSLATOR_TOOL.items()})
        assert tool is not None

        metadata_indexer.idx_storage.content_metadata_add([{
            'indexer_configuration_id': tool['id'],
            'id': b'cde',
            'metadata': YARN_PARSER_METADATA,
        }])

        sha1_gits = [
            hash_to_bytes('8dbb6aeb036e7fd80664eb8bfd1507881af1ba9f'),
        ]
        metadata_indexer.run(sha1_gits, 'update-dups')

        results = list(
            metadata_indexer.idx_storage.
            revision_intrinsic_metadata_get(sha1_gits))

        expected_results = [{
            'id': hash_to_bytes('8dbb6aeb036e7fd80664eb8bfd1507881af1ba9f'),
            'tool': TRANSLATOR_TOOL,
            'metadata': YARN_PARSER_METADATA,
            'mappings': ['npm'],
        }]

        for result in results:
            del result['tool']['id']

        # then
        self.assertEqual(expected_results, results)

    def test_revision_metadata_indexer_single_root_dir(self):
        metadata_indexer = RevisionMetadataIndexer(
            config=REVISION_METADATA_CONFIG)
        fill_obj_storage(metadata_indexer.objstorage)
        fill_storage(metadata_indexer.storage)

        # Add a parent directory, that is the only directory at the root
        # of the revision
        rev_id = hash_to_bytes('8dbb6aeb036e7fd80664eb8bfd1507881af1ba9f')
        rev = metadata_indexer.storage._revisions[rev_id]
        subdir_id = rev.directory
        rev = attr.evolve(rev, directory=b'123456')
        metadata_indexer.storage.directory_add([{
            'id': b'123456',
            'entries': [{
                'name': b'foobar-1.0.0',
                'type': 'dir',
                'target': subdir_id,
                'perms': 16384,
            }],
        }])

        tool = metadata_indexer.idx_storage.indexer_configuration_get(
            {'tool_'+k: v for (k, v) in TRANSLATOR_TOOL.items()})
        assert tool is not None

        metadata_indexer.idx_storage.content_metadata_add([{
            'indexer_configuration_id': tool['id'],
            'id': b'cde',
            'metadata': YARN_PARSER_METADATA,
        }])

        sha1_gits = [
            hash_to_bytes('8dbb6aeb036e7fd80664eb8bfd1507881af1ba9f'),
        ]
        metadata_indexer.run(sha1_gits, 'update-dups')

        results = list(
            metadata_indexer.idx_storage.
            revision_intrinsic_metadata_get(sha1_gits))

        expected_results = [{
            'id': hash_to_bytes('8dbb6aeb036e7fd80664eb8bfd1507881af1ba9f'),
            'tool': TRANSLATOR_TOOL,
            'metadata': YARN_PARSER_METADATA,
            'mappings': ['npm'],
        }]

        for result in results:
            del result['tool']['id']

        # then
        self.assertEqual(expected_results, results)



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest

from swh.indexer.origin_head import OriginHeadIndexer
from swh.indexer.tests.utils import (
    BASE_TEST_CONFIG, fill_storage
)

ORIGIN_HEAD_CONFIG = {
    **BASE_TEST_CONFIG,
    'tools': {
        'name': 'origin-metadata',
        'version': '0.0.1',
        'configuration': {},
    },
    'tasks': {
        'revision_intrinsic_metadata': None,
        'origin_intrinsic_metadata': None,
    }
}


class OriginHeadTestIndexer(OriginHeadIndexer):
    """Specific indexer whose configuration is enough to satisfy the
       indexing tests.
    """
    def parse_config_file(self, *args, **kwargs):
        return ORIGIN_HEAD_CONFIG

    def persist_index_computations(self, results, policy_update):
        self.results = results


class OriginHead(unittest.TestCase):
    def setUp(self):
        self.indexer = OriginHeadTestIndexer()
        self.indexer.catch_exceptions = False
        fill_storage(self.indexer.storage)

    def test_git(self):
        self.indexer.run(
                ['https://github.com/SoftwareHeritage/swh-storage'])
        self.assertEqual(self.indexer.results, [{
            'revision_id': b'8K\x12\x00d\x03\xcc\xe4]bS\xe3\x8f{'
                           b'\xd7}\xac\xefrm',
            'origin_url': 'https://github.com/SoftwareHeritage/swh-storage'}])

    def test_git_partial_snapshot(self):
        """Checks partial snapshots are ignored."""
        origin_url = 'https://github.com/SoftwareHeritage/swh-core'
        self.indexer.storage.origin_add_one({
            'url': origin_url,
        })
        visit = self.indexer.storage.origin_visit_add(
            origin_url, '2019-02-27', type='git',
        )
        self.indexer.storage.snapshot_add([{
            'id': b'foo',
            'branches': {
                b'foo': None,
                b'HEAD': {
                    'target_type': 'alias',
                    'target': b'foo',
                }
            }
        }])
        self.indexer.storage.origin_visit_update(
            origin_url, visit['visit'], status='partial', snapshot=b'foo')
        self.indexer.run([origin_url])
        self.assertEqual(self.indexer.results, [])

    def test_vcs_missing_snapshot(self):
        self.indexer.storage.origin_add([{
            'type': 'git',
            'url': 'https://github.com/SoftwareHeritage/swh-indexer',
        }])
        self.indexer.run(
                ['https://github.com/SoftwareHeritage/swh-indexer'])
        self.assertEqual(self.indexer.results, [])

    def test_pypi_missing_branch(self):
        origin_url = 'https://pypi.org/project/abcdef/'
        self.indexer.storage.origin_add_one({
            'url': origin_url,
        })
        visit = self.indexer.storage.origin_visit_add(
            origin_url, '2019-02-27', type='pypi')
        self.indexer.storage.snapshot_add([{
            'id': b'foo',
            'branches': {
                b'foo': None,
                b'HEAD': {
                    'target_type': 'alias',
                    'target': b'foo',
                }
            }
        }])
        self.indexer.storage.origin_visit_update(
            origin_url, visit['visit'], status='full', snapshot=b'foo')
        self.indexer.run(['https://pypi.org/project/abcdef/'])
        self.assertEqual(self.indexer.results, [])

    def test_ftp(self):
        self.indexer.run(
                ['rsync://ftp.gnu.org/gnu/3dldf'])
        self.assertEqual(self.indexer.results, [{
            'revision_id': b'\x8e\xa9\x8e/\xea}\x9feF\xf4\x9f\xfd\xee'
                           b'\xcc\x1a\xb4`\x8c\x8by',
            'origin_url': 'rsync://ftp.gnu.org/gnu/3dldf'}])

    def test_ftp_missing_snapshot(self):
        self.indexer.storage.origin_add([{
            'type': 'ftp',
            'url': 'rsync://ftp.gnu.org/gnu/foobar',
        }])
        self.indexer.run(
                ['rsync://ftp.gnu.org/gnu/foobar'])
        self.assertEqual(self.indexer.results, [])

    def test_deposit(self):
        self.indexer.run(
                ['https://forge.softwareheritage.org/source/'
                 'jesuisgpl/'])
        self.assertEqual(self.indexer.results, [{
            'revision_id': b'\xe7n\xa4\x9c\x9f\xfb\xb7\xf76\x11\x08{'
                           b'\xa6\xe9\x99\xb1\x9e]q\xeb',
            'origin_url': 'https://forge.softwareheritage.org/source/'
                          'jesuisgpl/'}])

    def test_deposit_missing_snapshot(self):
        self.indexer.storage.origin_add([{
            'type': 'deposit',
            'url': 'https://forge.softwareheritage.org/source/foobar',
        }])
        self.indexer.run(
                ['https://forge.softwareheritage.org/source/foobar'])
        self.assertEqual(self.indexer.results, [])

    def test_pypi(self):
        self.indexer.run(
                ['https://pypi.org/project/limnoria/'])
        self.assertEqual(self.indexer.results, [{
            'revision_id': b'\x83\xb9\xb6\xc7\x05\xb1%\xd0\xfem\xd8k'
                           b'A\x10\x9d\xc5\xfa2\xf8t',
            'origin_url': 'https://pypi.org/project/limnoria/'}])

    def test_svn(self):
        self.indexer.run(
                ['http://0-512-md.googlecode.com/svn/'])
        self.assertEqual(self.indexer.results, [{
            'revision_id': b'\xe4?r\xe1,\x88\xab\xec\xe7\x9a\x87\xb8'
                           b'\xc9\xad#.\x1bw=\x18',
            'origin_url': 'http://0-512-md.googlecode.com/svn/'}])



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest
import yaml

from swh.indexer.storage.api.server import load_and_check_config


def prepare_config_file(tmpdir, content, name='config.yml'):
    """Prepare configuration file in `$tmpdir/name` with content `content`.

    Args:
        tmpdir (LocalPath): root directory
        content (str/dict): Content of the file either as string or as a dict.
                            If a dict, converts the dict into a yaml string.
        name (str): configuration filename

    Returns
        path (str) of the configuration file prepared.

    """
    config_path = tmpdir / name
    if isinstance(content, dict):  # convert if needed
        content = yaml.dump(content)
    config_path.write_text(content, encoding='utf-8')
    # pytest on python3.5 does not support LocalPath manipulation, so
    # convert path to string
    return str(config_path)


def test_load_and_check_config_no_configuration():
    """Inexistent configuration files raises"""
    with pytest.raises(EnvironmentError) as e:
        load_and_check_config(None)

    assert e.value.args[0] == 'Configuration file must be defined'

    config_path = '/indexer/inexistent/config.yml'
    with pytest.raises(FileNotFoundError) as e:
        load_and_check_config(config_path)

    assert e.value.args[0] == 'Configuration file %s does not exist' % (
        config_path, )


def test_load_and_check_config_wrong_configuration(tmpdir):
    """Wrong configuration raises"""
    config_path = prepare_config_file(tmpdir, 'something: useless')
    with pytest.raises(KeyError) as e:
        load_and_check_config(config_path)

    assert e.value.args[0] == 'Missing \'%indexer_storage\' configuration'


def test_load_and_check_config_remote_config_local_type_raise(tmpdir):
    """'local' configuration without 'local' storage raises"""
    config = {
        'indexer_storage': {
            'cls': 'remote',
            'args': {}
        }
    }
    config_path = prepare_config_file(tmpdir, config)
    with pytest.raises(ValueError) as e:
        load_and_check_config(config_path, type='local')

    assert (
        e.value.args[0] ==
        "The indexer_storage backend can only be started with a 'local' "
        "configuration"
    )


def test_load_and_check_config_local_incomplete_configuration(tmpdir):
    """Incomplete 'local' configuration should raise"""
    config = {
        'indexer_storage': {
            'cls': 'local',
            'args': {
            }
        }
    }

    config_path = prepare_config_file(tmpdir, config)
    with pytest.raises(ValueError) as e:
        load_and_check_config(config_path)

    assert (
        e.value.args[0] ==
        "Invalid configuration; missing 'db' config entry"
    )


def test_load_and_check_config_local_config_fine(tmpdir):
    """'Remote configuration is fine"""
    config = {
        'indexer_storage': {
            'cls': 'local',
            'args': {
                'db': 'db',
            }
        }
    }
    config_path = prepare_config_file(tmpdir, config)
    cfg = load_and_check_config(config_path, type='local')
    assert cfg == config


def test_load_and_check_config_remote_config_fine(tmpdir):
    """'Remote configuration is fine"""
    config = {
        'indexer_storage': {
            'cls': 'remote',
            'args': {}
        }
    }
    config_path = prepare_config_file(tmpdir, config)
    cfg = load_and_check_config(config_path, type='any')

    assert cfg == config



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from os.path import join
import pytest

from . import SQL_DIR
from swh.storage.tests.conftest import postgresql_fact
from swh.indexer.storage import get_indexer_storage
from swh.model.hashutil import hash_to_bytes
from .generate_data_test import (
    MIMETYPE_OBJECTS, FOSSOLOGY_LICENSES, TOOLS
)


DUMP_FILES = join(SQL_DIR, '*.sql')


class DataObj(dict):
    def __getattr__(self, key):
        return self.__getitem__(key)

    def __setattr__(self, key, value):
        return self.__setitem__(key, value)


@pytest.fixture
def swh_indexer_storage_with_data(swh_indexer_storage):
    data = DataObj()
    tools = {
        tool['tool_name']: {
            'id': tool['id'],
            'name': tool['tool_name'],
            'version': tool['tool_version'],
            'configuration': tool['tool_configuration'],
            }
        for tool in swh_indexer_storage.indexer_configuration_add(TOOLS)}
    data.tools = tools
    data.sha1_1 = hash_to_bytes(
        '34973274ccef6ab4dfaaf86599792fa9c3fe4689')
    data.sha1_2 = hash_to_bytes(
        '61c2b3a30496d329e21af70dd2d7e097046d07b7')
    data.revision_id_1 = hash_to_bytes(
        '7026b7c1a2af56521e951c01ed20f255fa054238')
    data.revision_id_2 = hash_to_bytes(
        '7026b7c1a2af56521e9587659012345678904321')
    data.revision_id_3 = hash_to_bytes(
        '7026b7c1a2af56521e9587659012345678904320')
    data.origin_url_1 = 'file:///dev/0/zero'  # 44434341
    data.origin_url_2 = 'file:///dev/1/one'   # 44434342
    data.origin_url_3 = 'file:///dev/2/two'   # 54974445
    data.mimetypes = [
        {**mimetype_obj, 'indexer_configuration_id': tools['file']['id']}
        for mimetype_obj in MIMETYPE_OBJECTS
    ]
    swh_indexer_storage.content_mimetype_add(data.mimetypes)
    data.fossology_licenses = [
        {**fossology_obj, 'indexer_configuration_id': tools['nomos']['id']}
        for fossology_obj in FOSSOLOGY_LICENSES
    ]
    swh_indexer_storage._test_data = data

    return (swh_indexer_storage, data)


swh_indexer_storage_postgresql = postgresql_fact(
    'postgresql_proc', dump_files=DUMP_FILES)


@pytest.fixture
def swh_indexer_storage(swh_indexer_storage_postgresql):
    storage_config = {
        'cls': 'local',
        'args': {
            'db': swh_indexer_storage_postgresql.dsn,
        },
    }
    return get_indexer_storage(**storage_config)



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from functools import reduce
import re
import tempfile
from unittest.mock import patch

from click.testing import CliRunner

from swh.journal.tests.utils import FakeKafkaMessage, MockedKafkaConsumer
from swh.model.hashutil import hash_to_bytes

from swh.indexer.cli import cli


CLI_CONFIG = '''
scheduler:
    cls: foo
    args: {}
storage:
    cls: memory
indexer_storage:
    cls: memory
    args: {}
'''


def fill_idx_storage(idx_storage, nb_rows):
    tools = [
        {
            'tool_name': 'tool %d' % i,
            'tool_version': '0.0.1',
            'tool_configuration': {},
        }
        for i in range(2)
    ]
    tools = idx_storage.indexer_configuration_add(tools)

    origin_metadata = [
        {
            'id': 'file://dev/%04d' % origin_id,
            'from_revision': hash_to_bytes('abcd{:0>4}'.format(origin_id)),
            'indexer_configuration_id': tools[origin_id % 2]['id'],
            'metadata': {'name': 'origin %d' % origin_id},
            'mappings': ['mapping%d' % (origin_id % 10)]
        }
        for origin_id in range(nb_rows)
    ]
    revision_metadata = [
        {
            'id': hash_to_bytes('abcd{:0>4}'.format(origin_id)),
            'indexer_configuration_id': tools[origin_id % 2]['id'],
            'metadata': {'name': 'origin %d' % origin_id},
            'mappings': ['mapping%d' % (origin_id % 10)]
        }
        for origin_id in range(nb_rows)
    ]

    idx_storage.revision_intrinsic_metadata_add(revision_metadata)
    idx_storage.origin_intrinsic_metadata_add(origin_metadata)

    return [tool['id'] for tool in tools]


def _origins_in_task_args(tasks):
    """Returns the set of origins contained in the arguments of the
    provided tasks (assumed to be of type index-origin-metadata)."""
    return reduce(
        set.union,
        (set(task['arguments']['args'][0]) for task in tasks),
        set()
    )


def _assert_tasks_for_origins(tasks, origins):
    expected_kwargs = {"policy_update": "update-dups"}
    assert {task['type'] for task in tasks} == {'index-origin-metadata'}
    assert all(len(task['arguments']['args']) == 1 for task in tasks)
    for task in tasks:
        assert task['arguments']['kwargs'] == expected_kwargs, task
    assert _origins_in_task_args(tasks) == set([
        'file://dev/%04d' % i for i in origins])


def invoke(scheduler, catch_exceptions, args):
    runner = CliRunner()
    with patch('swh.indexer.cli.get_scheduler') as get_scheduler_mock, \
            tempfile.NamedTemporaryFile('a', suffix='.yml') as config_fd:
        config_fd.write(CLI_CONFIG)
        config_fd.seek(0)
        get_scheduler_mock.return_value = scheduler
        result = runner.invoke(cli, ['-C' + config_fd.name] + args)
    if not catch_exceptions and result.exception:
        print(result.output)
        raise result.exception
    return result


def test_mapping_list(indexer_scheduler):
    result = invoke(indexer_scheduler, False, [
        'mapping', 'list',
    ])
    expected_output = '\n'.join([
        'codemeta', 'gemspec', 'maven', 'npm', 'pkg-info', '',
    ])
    assert result.exit_code == 0, result.output
    assert result.output == expected_output


def test_mapping_list_terms(indexer_scheduler):
    result = invoke(indexer_scheduler, False, [
        'mapping', 'list-terms',
    ])
    assert result.exit_code == 0, result.output
    assert re.search(r'http://schema.org/url:\n.*npm', result.output)
    assert re.search(r'http://schema.org/url:\n.*codemeta', result.output)
    assert re.search(
        r'https://codemeta.github.io/terms/developmentStatus:\n\tcodemeta',
        result.output)


def test_mapping_list_terms_exclude(indexer_scheduler):
    result = invoke(indexer_scheduler, False, [
        'mapping', 'list-terms',
        '--exclude-mapping', 'codemeta'
    ])
    assert result.exit_code == 0, result.output
    assert re.search(r'http://schema.org/url:\n.*npm', result.output)
    assert not re.search(r'http://schema.org/url:\n.*codemeta', result.output)
    assert not re.search(
        r'https://codemeta.github.io/terms/developmentStatus:\n\tcodemeta',
        result.output)


@patch('swh.scheduler.cli.utils.TASK_BATCH_SIZE', 3)
@patch('swh.scheduler.cli_utils.TASK_BATCH_SIZE', 3)
def test_origin_metadata_reindex_empty_db(
        indexer_scheduler, idx_storage, storage):
    result = invoke(indexer_scheduler, False, [
        'schedule', 'reindex_origin_metadata',
    ])
    expected_output = (
        'Nothing to do (no origin metadata matched the criteria).\n'
    )
    assert result.exit_code == 0, result.output
    assert result.output == expected_output
    tasks = indexer_scheduler.search_tasks()
    assert len(tasks) == 0


@patch('swh.scheduler.cli.utils.TASK_BATCH_SIZE', 3)
@patch('swh.scheduler.cli_utils.TASK_BATCH_SIZE', 3)
def test_origin_metadata_reindex_divisor(
        indexer_scheduler, idx_storage, storage):
    """Tests the re-indexing when origin_batch_size*task_batch_size is a
    divisor of nb_origins."""
    fill_idx_storage(idx_storage, 90)

    result = invoke(indexer_scheduler, False, [
        'schedule', 'reindex_origin_metadata',
    ])

    # Check the output
    expected_output = (
        'Scheduled 3 tasks (30 origins).\n'
        'Scheduled 6 tasks (60 origins).\n'
        'Scheduled 9 tasks (90 origins).\n'
        'Done.\n'
    )
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    # Check scheduled tasks
    tasks = indexer_scheduler.search_tasks()
    assert len(tasks) == 9
    _assert_tasks_for_origins(tasks, range(90))


@patch('swh.scheduler.cli.utils.TASK_BATCH_SIZE', 3)
@patch('swh.scheduler.cli_utils.TASK_BATCH_SIZE', 3)
def test_origin_metadata_reindex_dry_run(
        indexer_scheduler, idx_storage, storage):
    """Tests the re-indexing when origin_batch_size*task_batch_size is a
    divisor of nb_origins."""
    fill_idx_storage(idx_storage, 90)

    result = invoke(indexer_scheduler, False, [
        'schedule', '--dry-run', 'reindex_origin_metadata',
    ])

    # Check the output
    expected_output = (
        'Scheduled 3 tasks (30 origins).\n'
        'Scheduled 6 tasks (60 origins).\n'
        'Scheduled 9 tasks (90 origins).\n'
        'Done.\n'
    )
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    # Check scheduled tasks
    tasks = indexer_scheduler.search_tasks()
    assert len(tasks) == 0


@patch('swh.scheduler.cli.utils.TASK_BATCH_SIZE', 3)
@patch('swh.scheduler.cli_utils.TASK_BATCH_SIZE', 3)
def test_origin_metadata_reindex_nondivisor(
        indexer_scheduler, idx_storage, storage):
    """Tests the re-indexing when neither origin_batch_size or
    task_batch_size is a divisor of nb_origins."""
    fill_idx_storage(idx_storage, 70)

    result = invoke(indexer_scheduler, False, [
        'schedule', 'reindex_origin_metadata',
        '--batch-size', '20',
    ])

    # Check the output
    expected_output = (
        'Scheduled 3 tasks (60 origins).\n'
        'Scheduled 4 tasks (70 origins).\n'
        'Done.\n'
    )
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    # Check scheduled tasks
    tasks = indexer_scheduler.search_tasks()
    assert len(tasks) == 4
    _assert_tasks_for_origins(tasks, range(70))


@patch('swh.scheduler.cli.utils.TASK_BATCH_SIZE', 3)
@patch('swh.scheduler.cli_utils.TASK_BATCH_SIZE', 3)
def test_origin_metadata_reindex_filter_one_mapping(
        indexer_scheduler, idx_storage, storage):
    """Tests the re-indexing when origin_batch_size*task_batch_size is a
    divisor of nb_origins."""
    fill_idx_storage(idx_storage, 110)

    result = invoke(indexer_scheduler, False, [
        'schedule', 'reindex_origin_metadata',
        '--mapping', 'mapping1',
    ])

    # Check the output
    expected_output = (
        'Scheduled 2 tasks (11 origins).\n'
        'Done.\n'
    )
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    # Check scheduled tasks
    tasks = indexer_scheduler.search_tasks()
    assert len(tasks) == 2
    _assert_tasks_for_origins(
        tasks,
        [1, 11, 21, 31, 41, 51, 61, 71, 81, 91, 101])


@patch('swh.scheduler.cli.utils.TASK_BATCH_SIZE', 3)
@patch('swh.scheduler.cli_utils.TASK_BATCH_SIZE', 3)
def test_origin_metadata_reindex_filter_two_mappings(
        indexer_scheduler, idx_storage, storage):
    """Tests the re-indexing when origin_batch_size*task_batch_size is a
    divisor of nb_origins."""
    fill_idx_storage(idx_storage, 110)

    result = invoke(indexer_scheduler, False, [
        'schedule', 'reindex_origin_metadata',
        '--mapping', 'mapping1', '--mapping', 'mapping2',
    ])

    # Check the output
    expected_output = (
        'Scheduled 3 tasks (22 origins).\n'
        'Done.\n'
    )
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    # Check scheduled tasks
    tasks = indexer_scheduler.search_tasks()
    assert len(tasks) == 3
    _assert_tasks_for_origins(
        tasks,
        [1, 11, 21, 31, 41, 51, 61, 71, 81, 91, 101,
         2, 12, 22, 32, 42, 52, 62, 72, 82, 92, 102])


@patch('swh.scheduler.cli.utils.TASK_BATCH_SIZE', 3)
@patch('swh.scheduler.cli_utils.TASK_BATCH_SIZE', 3)
def test_origin_metadata_reindex_filter_one_tool(
        indexer_scheduler, idx_storage, storage):
    """Tests the re-indexing when origin_batch_size*task_batch_size is a
    divisor of nb_origins."""
    tool_ids = fill_idx_storage(idx_storage, 110)

    result = invoke(indexer_scheduler, False, [
        'schedule', 'reindex_origin_metadata',
        '--tool-id', str(tool_ids[0]),
    ])

    # Check the output
    expected_output = (
        'Scheduled 3 tasks (30 origins).\n'
        'Scheduled 6 tasks (55 origins).\n'
        'Done.\n'
    )
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    # Check scheduled tasks
    tasks = indexer_scheduler.search_tasks()
    assert len(tasks) == 6
    _assert_tasks_for_origins(
        tasks,
        [x*2 for x in range(55)])


def test_journal_client(storage, indexer_scheduler):
    """Test the 'swh indexer journal-client' cli tool."""
    message = FakeKafkaMessage('swh.journal.objects.origin_visit', 'bogus', {
        'status': 'full',
        'origin': {
            'url': 'file://dev/0000',
        }
    })

    consumer = MockedKafkaConsumer([message])

    with patch('swh.journal.client.Consumer',
               return_value=consumer):
        result = invoke(indexer_scheduler, False, [
            'journal-client',
            '--max-messages', '1',
            '--broker', '192.0.2.1',
            '--prefix', 'swh.journal.objects',
            '--group-id', 'test-consumer',
        ])

    # Check the output
    expected_output = (
        'Processed 1 messages.\n'
        'Done.\n'
    )
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    # Check scheduled tasks
    tasks = indexer_scheduler.search_tasks()
    assert len(tasks) == 1
    _assert_tasks_for_origins(
        tasks,
        [0])



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import functools
import json
import time

import click

from swh.core import config
from swh.core.cli import CONTEXT_SETTINGS, AliasedGroup
from swh.journal.cli import get_journal_client
from swh.scheduler import get_scheduler
from swh.scheduler.cli_utils import schedule_origin_batches
from swh.storage import get_storage

from swh.indexer import metadata_dictionary
from swh.indexer.journal_client import process_journal_objects
from swh.indexer.storage import get_indexer_storage
from swh.indexer.storage.api.server import load_and_check_config, app


@click.group(name='indexer', context_settings=CONTEXT_SETTINGS,
             cls=AliasedGroup)
@click.option('--config-file', '-C', default=None,
              type=click.Path(exists=True, dir_okay=False,),
              help="Configuration file.")
@click.pass_context
def cli(ctx, config_file):
    """Software Heritage Indexer tools.

    The Indexer is used to mine the content of the archive and extract derived
    information from archive source code artifacts.

    """
    ctx.ensure_object(dict)

    conf = config.read(config_file)
    ctx.obj['config'] = conf


def _get_api(getter, config, config_key, url):
    if url:
        config[config_key] = {
            'cls': 'remote',
            'args': {'url': url}
        }
    elif config_key not in config:
        raise click.ClickException(
            'Missing configuration for {}'.format(config_key))
    return getter(**config[config_key])


@cli.group('mapping')
def mapping():
    '''Manage Software Heritage Indexer mappings.'''
    pass


@mapping.command('list')
def mapping_list():
    """Prints the list of known mappings."""
    mapping_names = [mapping.name
                     for mapping in metadata_dictionary.MAPPINGS.values()]
    mapping_names.sort()
    for mapping_name in mapping_names:
        click.echo(mapping_name)


@mapping.command('list-terms')
@click.option('--exclude-mapping', multiple=True,
              help='Exclude the given mapping from the output')
@click.option('--concise', is_flag=True,
              default=False,
              help='Don\'t print the list of mappings supporting each term.')
def mapping_list_terms(concise, exclude_mapping):
    """Prints the list of known CodeMeta terms, and which mappings
    support them."""
    properties = metadata_dictionary.list_terms()
    for (property_name, supported_mappings) in sorted(properties.items()):
        supported_mappings = {m.name for m in supported_mappings}
        supported_mappings -= set(exclude_mapping)
        if supported_mappings:
            if concise:
                click.echo(property_name)
            else:
                click.echo('{}:'.format(property_name))
                click.echo('\t' + ', '.join(sorted(supported_mappings)))


@mapping.command('translate')
@click.argument('mapping-name')
@click.argument('file', type=click.File('rb'))
def mapping_translate(mapping_name, file):
    """Prints the list of known mappings."""
    mapping_cls = [cls for cls in metadata_dictionary.MAPPINGS.values()
                   if cls.name == mapping_name]
    if not mapping_cls:
        raise click.ClickException('Unknown mapping {}'.format(mapping_name))
    assert len(mapping_cls) == 1
    mapping_cls = mapping_cls[0]
    mapping = mapping_cls()
    codemeta_doc = mapping.translate(file.read())
    click.echo(json.dumps(codemeta_doc, indent=4))


@cli.group('schedule')
@click.option('--scheduler-url', '-s', default=None,
              help="URL of the scheduler API")
@click.option('--indexer-storage-url', '-i', default=None,
              help="URL of the indexer storage API")
@click.option('--storage-url', '-g', default=None,
              help="URL of the (graph) storage API")
@click.option('--dry-run/--no-dry-run', is_flag=True,
              default=False,
              help='List only what would be scheduled.')
@click.pass_context
def schedule(ctx, scheduler_url, storage_url, indexer_storage_url,
             dry_run):
    """Manipulate Software Heritage Indexer tasks.

    Via SWH Scheduler's API."""
    ctx.obj['indexer_storage'] = _get_api(
        get_indexer_storage,
        ctx.obj['config'],
        'indexer_storage',
        indexer_storage_url
    )
    ctx.obj['storage'] = _get_api(
        get_storage,
        ctx.obj['config'],
        'storage',
        storage_url
    )
    ctx.obj['scheduler'] = _get_api(
        get_scheduler,
        ctx.obj['config'],
        'scheduler',
        scheduler_url
    )
    if dry_run:
        ctx.obj['scheduler'] = None


def list_origins_by_producer(idx_storage, mappings, tool_ids):
    next_page_token = ''
    limit = 10000
    while next_page_token is not None:
        result = idx_storage.origin_intrinsic_metadata_search_by_producer(
            page_token=next_page_token, limit=limit, ids_only=True,
            mappings=mappings or None, tool_ids=tool_ids or None)
        next_page_token = result.get('next_page_token')
        yield from result['origins']


@schedule.command('reindex_origin_metadata')
@click.option('--batch-size', '-b', 'origin_batch_size',
              default=10, show_default=True, type=int,
              help="Number of origins per task")
@click.option('--tool-id', '-t', 'tool_ids', type=int, multiple=True,
              help="Restrict search of old metadata to this/these tool ids.")
@click.option('--mapping', '-m', 'mappings', multiple=True,
              help="Mapping(s) that should be re-scheduled (eg. 'npm', "
                   "'gemspec', 'maven')")
@click.option('--task-type',
              default='index-origin-metadata', show_default=True,
              help="Name of the task type to schedule.")
@click.pass_context
def schedule_origin_metadata_reindex(
        ctx, origin_batch_size, tool_ids, mappings, task_type):
    """Schedules indexing tasks for origins that were already indexed."""
    idx_storage = ctx.obj['indexer_storage']
    scheduler = ctx.obj['scheduler']

    origins = list_origins_by_producer(idx_storage, mappings, tool_ids)

    kwargs = {"policy_update": "update-dups"}
    schedule_origin_batches(
        scheduler, task_type, origins, origin_batch_size, kwargs)


@cli.command('journal-client')
@click.option('--scheduler-url', '-s', default=None,
              help="URL of the scheduler API")
@click.option('--origin-metadata-task-type',
              default='index-origin-metadata',
              help='Name of the task running the origin metadata indexer.')
@click.option('--broker', 'brokers', type=str, multiple=True,
              help='Kafka broker to connect to.')
@click.option('--prefix', type=str, default=None,
              help='Prefix of Kafka topic names to read from.')
@click.option('--group-id', type=str,
              help='Consumer/group id for reading from Kafka.')
@click.option('--max-messages', '-m', default=None, type=int,
              help='Maximum number of objects to replay. Default is to '
                   'run forever.')
@click.pass_context
def journal_client(ctx, scheduler_url, origin_metadata_task_type,
                   brokers, prefix, group_id, max_messages):
    """Listens for new objects from the SWH Journal, and schedules tasks
    to run relevant indexers (currently, only origin-intrinsic-metadata)
    on these new objects."""
    scheduler = _get_api(
        get_scheduler,
        ctx.obj['config'],
        'scheduler',
        scheduler_url
    )

    client = get_journal_client(
        ctx, brokers=brokers, prefix=prefix, group_id=group_id,
        object_types=['origin_visit'],
        max_messages=max_messages,
    )

    worker_fn = functools.partial(
        process_journal_objects,
        scheduler=scheduler,
        task_names={
            'origin_metadata': origin_metadata_task_type,
        }
    )
    nb_messages = 0
    last_log_time = 0
    try:
        while not max_messages or nb_messages < max_messages:
            nb_messages += client.process(worker_fn)
            if time.monotonic() - last_log_time >= 60:
                print('Processed %d messages.' % nb_messages)
                last_log_time = time.monotonic()
    except KeyboardInterrupt:
        ctx.exit(0)
    else:
        print('Done.')


@cli.command('rpc-serve')
@click.argument('config-path', required=True)
@click.option('--host', default='0.0.0.0', help="Host to run the server")
@click.option('--port', default=5007, type=click.INT,
              help="Binding port of the server")
@click.option('--debug/--nodebug', default=True,
              help="Indicates if the server should run in debug mode")
def rpc_server(config_path, host, port, debug):
    """Starts a Software Heritage Indexer RPC HTTP server."""
    api_cfg = load_and_check_config(config_path, type='any')
    app.config.update(api_cfg)
    app.run(host, port=int(port), debug=bool(debug))


def main():
    return cli(auto_envvar_prefix='SWH_INDEXER')


if __name__ == '__main__':
    main()



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import ast
import itertools
import re

from swh.indexer.codemeta import CROSSWALK_TABLE, SCHEMA_URI
from .base import DictMapping


def name_to_person(name):
    return {
        '@type': SCHEMA_URI + 'Person',
        SCHEMA_URI + 'name': name,
    }


class GemspecMapping(DictMapping):
    name = 'gemspec'
    mapping = CROSSWALK_TABLE['Ruby Gem']
    string_fields = ['name', 'version', 'description', 'summary', 'email']

    _re_spec_new = re.compile(r'.*Gem::Specification.new +(do|\{) +\|.*\|.*')
    _re_spec_entry = re.compile(r'\s*\w+\.(?P<key>\w+)\s*=\s*(?P<expr>.*)')

    @classmethod
    def detect_metadata_files(cls, file_entries):
        for entry in file_entries:
            if entry['name'].endswith(b'.gemspec'):
                return [entry['sha1']]
        return []

    def translate(self, raw_content):
        try:
            raw_content = raw_content.decode()
        except UnicodeDecodeError:
            self.log.warning('Error unidecoding from %s', self.log_suffix)
            return

        # Skip lines before 'Gem::Specification.new'
        lines = itertools.dropwhile(
            lambda x: not self._re_spec_new.match(x),
            raw_content.split('\n'))

        try:
            next(lines)  # Consume 'Gem::Specification.new'
        except StopIteration:
            self.log.warning('Could not find Gem::Specification in %s',
                             self.log_suffix)
            return

        content_dict = {}
        for line in lines:
            match = self._re_spec_entry.match(line)
            if match:
                value = self.eval_ruby_expression(match.group('expr'))
                if value:
                    content_dict[match.group('key')] = value
        return self._translate_dict(content_dict)

    def eval_ruby_expression(self, expr):
        """Very simple evaluator of Ruby expressions.

        >>> GemspecMapping().eval_ruby_expression('"Foo bar"')
        'Foo bar'
        >>> GemspecMapping().eval_ruby_expression("'Foo bar'")
        'Foo bar'
        >>> GemspecMapping().eval_ruby_expression("['Foo', 'bar']")
        ['Foo', 'bar']
        >>> GemspecMapping().eval_ruby_expression("'Foo bar'.freeze")
        'Foo bar'
        >>> GemspecMapping().eval_ruby_expression( \
                "['Foo'.freeze, 'bar'.freeze]")
        ['Foo', 'bar']
        """
        def evaluator(node):
            if isinstance(node, ast.Str):
                return node.s
            elif isinstance(node, ast.List):
                res = []
                for element in node.elts:
                    val = evaluator(element)
                    if not val:
                        return
                    res.append(val)
                return res

        expr = expr.replace('.freeze', '')
        try:
            # We're parsing Ruby expressions here, but Python's
            # ast.parse works for very simple Ruby expressions
            # (mainly strings delimited with " or ', and lists
            # of such strings).
            tree = ast.parse(expr, mode='eval')
        except (SyntaxError, ValueError):
            return
        if isinstance(tree, ast.Expression):
            return evaluator(tree.body)

    def normalize_homepage(self, s):
        if isinstance(s, str):
            return {"@id": s}

    def normalize_license(self, s):
        if isinstance(s, str):
            return [{"@id": "https://spdx.org/licenses/" + s}]

    def normalize_licenses(self, licenses):
        if isinstance(licenses, list):
            return [{"@id": "https://spdx.org/licenses/" + license}
                    for license in licenses
                    if isinstance(license, str)]

    def normalize_author(self, author):
        if isinstance(author, str):
            return {"@list": [name_to_person(author)]}

    def normalize_authors(self, authors):
        if isinstance(authors, list):
            return {"@list": [name_to_person(author) for author in authors
                              if isinstance(author, str)]}



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from uuid import uuid1

from swh.model.hashutil import MultiHash
from hypothesis.strategies import (composite, sets, one_of, uuids,
                                   tuples, sampled_from)
MIMETYPES = [
    b'application/json',
    b'application/octet-stream',
    b'application/xml',
    b'text/plain',
]

ENCODINGS = [
    b'iso8859-1',
    b'iso8859-15',
    b'latin1',
    b'utf-8',
]


def gen_mimetype():
    """Generate one mimetype strategy.

    """
    return one_of(sampled_from(MIMETYPES))


def gen_encoding():
    """Generate one encoding strategy.

    """
    return one_of(sampled_from(ENCODINGS))


def _init_content(uuid):
    """Given a uuid, initialize a content

    """
    return {
        'id': MultiHash.from_data(uuid.bytes, {'sha1'}).digest()['sha1'],
        'indexer_configuration_id': 1,
    }


@composite
def gen_content_mimetypes(draw, *, min_size=0, max_size=100):
    """Generate valid and consistent content_mimetypes.

    Context: Test purposes

    Args:
        **draw** (callable): Used by hypothesis to generate data
        **min_size** (int): Minimal number of elements to generate
                            (default: 0)
        **max_size** (int): Maximal number of elements to generate
                            (default: 100)

    Returns:
        List of content_mimetypes as expected by the
        content_mimetype_add api endpoint.

    """
    _ids = draw(
        sets(
            tuples(
                uuids(),
                gen_mimetype(),
                gen_encoding()
            ),
            min_size=min_size, max_size=max_size
        )
    )

    content_mimetypes = []
    for uuid, mimetype, encoding in _ids:
        content_mimetypes.append({
            **_init_content(uuid),
            'mimetype': mimetype,
            'encoding': encoding,
        })
    return content_mimetypes


TOOLS = [
    {
        'tool_name': 'universal-ctags',
        'tool_version': '~git7859817b',
        'tool_configuration': {
            "command_line": "ctags --fields=+lnz --sort=no --links=no "
                            "--output-format=json <filepath>"}
    },
    {
        'tool_name': 'swh-metadata-translator',
        'tool_version': '0.0.1',
        'tool_configuration': {"type": "local", "context": "NpmMapping"},
    },
    {
        'tool_name': 'swh-metadata-detector',
        'tool_version': '0.0.1',
        'tool_configuration': {
            "type": "local", "context": ["NpmMapping", "CodemetaMapping"]},
    },
    {
        'tool_name': 'swh-metadata-detector2',
        'tool_version': '0.0.1',
        'tool_configuration': {
            "type": "local", "context": ["NpmMapping", "CodemetaMapping"]},
    },
    {
        'tool_name': 'file',
        'tool_version': '5.22',
        'tool_configuration': {"command_line": "file --mime <filepath>"},
    },
    {
        'tool_name': 'pygments',
        'tool_version': '2.0.1+dfsg-1.1+deb8u1',
        'tool_configuration': {
            "type": "library", "debian-package": "python3-pygments"},
    },
    {
        'tool_name': 'pygments2',
        'tool_version': '2.0.1+dfsg-1.1+deb8u1',
        'tool_configuration': {
            "type": "library",
            "debian-package": "python3-pygments",
            "max_content_size": 10240
        },
    },
    {
        'tool_name': 'nomos',
        'tool_version': '3.1.0rc2-31-ga2cbb8c',
        'tool_configuration': {"command_line": "nomossa <filepath>"},
    },
]


MIMETYPE_OBJECTS = [
    {'id': MultiHash.from_data(uuid1().bytes, {'sha1'}).digest()['sha1'],
     'mimetype': mt,
     'encoding': enc,
     # 'indexer_configuration_id' will be added after TOOLS get registered
     }
    for mt in MIMETYPES
    for enc in ENCODINGS]

LICENSES = [
    b'3DFX',
    b'BSD',
    b'GPL',
    b'Apache2',
    b'MIT',
]

FOSSOLOGY_LICENSES = [
    {'id': MultiHash.from_data(uuid1().bytes, {'sha1'}).digest()['sha1'],
     'licenses': [LICENSES[i % len(LICENSES)], ],
     # 'indexer_configuration_id' will be added after TOOLS get registered
     }
    for i in range(10)
    ]


def gen_license():
    return one_of(sampled_from(LICENSES))


@composite
def gen_content_fossology_licenses(draw, *, min_size=0, max_size=100):
    """Generate valid and consistent content_fossology_licenses.

    Context: Test purposes

    Args:
        **draw** (callable): Used by hypothesis to generate data
        **min_size** (int): Minimal number of elements to generate
                            (default: 0)
        **max_size** (int): Maximal number of elements to generate
                            (default: 100)

    Returns:
        List of content_fossology_licenses as expected by the
        content_fossology_license_add api endpoint.

    """
    _ids = draw(
        sets(
            tuples(
                uuids(),
                gen_license(),
            ),
            min_size=min_size, max_size=max_size
        )
    )

    content_licenses = []
    for uuid, license in _ids:
        content_licenses.append({
            **_init_content(uuid),
            'licenses': [license],
        })
    return content_licenses
